"""

>>> from blaze.expr import symbol
>>> from blaze.compute.pandas import compute

>>> accounts = symbol('accounts', 'var * {name: string, amount: int}')
>>> deadbeats = accounts[accounts['amount'] < 0]['name']

>>> from pandas import DataFrame
>>> data = [['Alice', 100], ['Bob', -50], ['Charlie', -20]]
>>> df = DataFrame(data, columns=['name', 'amount'])
>>> compute(deadbeats, df)
1        Bob
2    Charlie
Name: name, dtype: object
"""
from __future__ import absolute_import, division, print_function

from datetime import timedelta
import fnmatch
import itertools
from distutils.version import LooseVersion
import warnings
from collections import defaultdict

import numpy as np

import pandas as pd

from pandas.core.generic import NDFrame
from pandas import DataFrame, Series
from pandas.core.groupby import DataFrameGroupBy, SeriesGroupBy

from toolz import merge as merge_dicts
from toolz import groupby
from toolz.curried import pipe, filter, map, concat

import datashape

from datashape import to_numpy_dtype
from datashape.predicates import isscalar

from odo import into
try:
    import dask.dataframe as dd
    DaskDataFrame = dd.DataFrame
    DaskSeries = dd.Series
except ImportError:
    DaskDataFrame = pd.DataFrame
    DaskSeries = pd.Series

from .core import compute, compute_up, base
from ..compatibility import _inttypes
from ..dispatch import dispatch
from ..expr import (
    Apply,
    BinOp,
    Broadcast,
    By,
    Ceil,
    Coalesce,
    Coerce,
    Concat,
    DateTime,
    DateTimeTruncate,
    Distinct,
    ElemWise,
    Expr,
    Field,
    Floor,
    Head,
    Interp,
    IsIn,
    Join,
    Label,
    Like,
    Map,
    Merge,
    Millisecond,
    Pad,
    Projection,
    ReLabel,
    Reduction,
    Replace,
    Round,
    Sample,
    seconds,
    Selection,
    Shift,
    Slice,
    Sort,
    strftime,
    Summary,
    Tail,
    UTCFromTimestamp,
    UnaryOp,
    UnaryStringFunction,
    common_subexpression,
    count,
    isnan,
    nelements,
    notnull,
    nunique,
    std,
    summary,
    symbol,
    var,
    StrCat,
    StrFind,
    StrSlice,
    SliceReplace,
    total_seconds,
)

__all__ = []


@dispatch(Projection, (DataFrame, DaskDataFrame))
def compute_up(t, df, **kwargs):
    return df[list(t.fields)]


@dispatch(Field, (DataFrame, DataFrameGroupBy, DaskDataFrame))
def compute_up(t, df, **kwargs):
    assert len(t.fields) == 1
    return df[t.fields[0]]


@dispatch(Field, (Series, DaskSeries))
def compute_up(t, data, **kwargs):
    assert len(t.fields) == 1
    if t.fields[0] == data.name:
        return data
    else:
        raise ValueError("Fieldname %r does not match Series name %r"
                         % (t.fields[0], data.name))


@dispatch(Broadcast, (DataFrame, DaskDataFrame))
def compute_up(t, df, **kwargs):
    return compute(t._full_expr, df, return_type='native')


@dispatch(Broadcast, Series)
def compute_up(t, s, **kwargs):
    return compute_up(t, s.to_frame(), **kwargs)


@dispatch(Interp, Series)
def compute_up(t, data, **kwargs):
    if isinstance(t.lhs, Expr):
        return data % t.rhs
    else:
        return t.lhs % data


@compute_up.register(Interp, Series, (Series, base))
@compute_up.register(Interp, base, Series)
def compute_up_pd_interp(t, lhs, rhs, **kwargs):
    return lhs % rhs



@dispatch(BinOp, (Series, DaskSeries))
def compute_up(t, data, **kwargs):
    if isinstance(t.lhs, Expr):
        return t.op(data, t.rhs)
    else:
        return t.op(t.lhs, data)

@compute_up.register(BinOp, (Series, DaskSeries), (Series, base, DaskSeries))
@compute_up.register(BinOp, (Series, base, DaskSeries), (Series, DaskSeries))
def compute_up_binop(t, lhs, rhs, **kwargs):
    return t.op(lhs, rhs)


@dispatch(UnaryOp, NDFrame)
def compute_up(t, df, **kwargs):
    f = getattr(t, 'op', getattr(np, t.symbol, None))
    if f is None:
        raise ValueError('%s is not a valid operation on %s objects' %
                         (t.symbol, type(df).__name__))
    return f(df)


@dispatch(Selection, (Series, DataFrame, DaskSeries, DaskDataFrame))
def compute_up(expr, df, **kwargs):
    return compute_up(
        expr,
        df,
        compute(expr.predicate, {expr._child: df}, return_type='native'),
        **kwargs
    )


@dispatch(Selection, (Series, DataFrame, DaskSeries, DaskDataFrame),
                     (Series, DaskSeries))
def compute_up(expr, df, predicate, **kwargs):
    return df[predicate]


@dispatch(Join, DataFrame, DataFrame)
def compute_up(t, lhs, rhs, **kwargs):
    """ Join two pandas data frames on arbitrary columns

    The approach taken here could probably be improved.

    To join on two columns we force each column to be the index of the
    dataframe, perform the join, and then reset the index back to the left
    side's original index.
    """
    result = pd.merge(
        lhs,
        rhs,
        left_on=t.on_left,
        right_on=t.on_right,
        how=t.how,
        suffixes=t.suffixes,
    )
    return result.reset_index()[t.fields]


@dispatch(isnan, pd.Series)
def compute_up(expr, data, **kwargs):
    return data.isnull()


@dispatch(notnull, pd.Series)
def compute_up(expr, data, **kwargs):
    return data.notnull()


pandas_structure = DataFrame, DaskDataFrame, Series, DataFrameGroupBy, SeriesGroupBy


@dispatch(Concat, pandas_structure, pandas_structure)
def compute_up(t, lhs, rhs, _concat=pd.concat, **kwargs):
    if not (isinstance(lhs, type(rhs)) or isinstance(rhs, type(lhs))):
        raise TypeError('lhs and rhs must be the same type')

    return _concat((lhs, rhs), axis=t.axis, ignore_index=True)


def get_scalar(result):
    # pandas may return an int, numpy scalar or non scalar here so we need to
    # program defensively so that things are JSON serializable
    try:
        return result.item()
    except (AttributeError, ValueError):
        return result


@dispatch(Reduction, (Series, SeriesGroupBy, DaskSeries))
def compute_up(t, s, **kwargs):
    result = get_scalar(getattr(s, t.symbol)())
    if t.keepdims:
        result = Series([result], name=s.name)
    return result


@dispatch((std, var), (Series, SeriesGroupBy))
def compute_up(t, s, **kwargs):
    measure = t.schema.measure
    is_timedelta = isinstance(
        getattr(measure, 'ty', measure),
        datashape.TimeDelta,
    )
    if is_timedelta:
        # part 1 of 2 to work around the fact that pandas does not have
        # timedelta var or std: cast to a double
        s = s.astype('timedelta64[s]').astype('int64')
    result = get_scalar(getattr(s, t.symbol)(ddof=t.unbiased))
    if t.keepdims:
        result = Series([result], name=s.name)
    if is_timedelta:
        # part 2 of 2 to work around the fact that postgres does not have
        # timedelta var or std: cast back from seconds by creating a timedelta
        result = timedelta(seconds=result)
    return result


@dispatch(Distinct, DataFrame)
def compute_up(t, df, **kwargs):
    return df.drop_duplicates(subset=t.on or None).reset_index(drop=True)


@dispatch(Distinct, Series)
def compute_up(t, s, **kwargs):
    if t.on:
        raise ValueError('malformed expression: no columns to distinct on')
    return s.drop_duplicates().reset_index(drop=True)


@dispatch(nunique, DataFrame)
def compute_up(expr, data, **kwargs):
    return compute_up(expr._child.distinct().count(), data, **kwargs)


@dispatch(UnaryStringFunction, Series)
def compute_up(expr, data, **kwargs):
    name = type(expr).__name__
    return getattr(data.str, name)()

@dispatch(Replace, Series)
def compute_up(expr, data, **kwargs):
    max = expr.max is None and -1 or expr.max
    return data.str.replace(expr.old, expr.new, max)

@dispatch(Pad, Series)
def compute_up(expr, data, **kwargs):
    side = expr.side is None and 'left' or expr.side
    fillchar = expr.fillchar is None and ' ' or expr.fillchar
    return data.str.pad(expr.width, side, fillchar)

@dispatch(StrCat, Series, Series)
def compute_up(expr, lhs_data, rhs_data, **kwargs):
    res = lhs_data.str.cat(rhs_data, sep=expr.sep)
    return res


@dispatch(StrFind, Series)
def compute_up(expr, data, **kwargs):
    return data.str.find(expr.sub)


@dispatch(StrSlice, Series)
def compute_up(expr, data, **kwargs):
    if isinstance(expr.slice, tuple):
        return data.str[slice(*expr.slice)]
    return data.str[expr.slice]

@dispatch(SliceReplace, Series)
def compute_up(expr, data, **kwargs):
    return data.str.slice_replace(expr.start, expr.stop, expr.repl)

def unpack(seq):
    """ Unpack sequence of length one

    >>> unpack([1, 2, 3])
    [1, 2, 3]

    >>> unpack([1])
    1
    """
    seq = list(seq)
    if len(seq) == 1:
        seq = seq[0]
    return seq


Grouper = ElemWise, Series, list


@dispatch(By, list, DataFrame)
def get_grouper(c, grouper, df):
    return grouper


@dispatch(By, Expr, NDFrame)
def get_grouper(c, grouper, df):
    g = compute(grouper, {c._child: df}, return_type='native')
    if isinstance(g, Series):
        return g
    if isinstance(g, DataFrame):
        return [g[col] for col in g.columns]


@dispatch(By, (Field, Projection), NDFrame)
def get_grouper(c, grouper, df):
    return grouper.fields


@dispatch(By, Reduction, Grouper, NDFrame)
def compute_by(t, r, g, df):
    names = [r._name]
    preapply = compute(r._child, {t._child: df}, return_type='native')

    # Pandas and Blaze column naming schemes differ
    # Coerce DataFrame column names to match Blaze's names
    preapply = preapply.copy()
    if isinstance(preapply, Series):
        preapply.name = names[0]
    else:
        preapply.names = names
    group_df = concat_nodup(df, preapply)

    gb = group_df.groupby(g)
    groups = gb[names[0] if isscalar(t.apply._child.dshape.measure) else names]

    return compute_up(r, groups, return_type='native')  # do reduction


name_dict = dict()
seen_names = set()


def _name(expr):
    """ A unique and deterministic name for an expression """
    if expr in name_dict:
        return name_dict[expr]
    result = base = expr._name or '_'
    if result in seen_names:
        for i in itertools.count(1):
            result = '%s_%d' % (base, i)
            if result not in seen_names:
                break
    # result is an unseen name
    seen_names.add(result)
    name_dict[expr] = result
    return result


def fancify_summary(expr):
    """ Separate a complex summary into two pieces

    Helps pandas compute_by on summaries

    >>> t = symbol('t', 'var * {x: int, y: int}')
    >>> one, two, three = fancify_summary(summary(a=t.x.sum(), b=t.x.sum() + t.y.count() - 1))

    A simpler summary with only raw reductions
    >>> one
    summary(x_sum=sum(t.x), y_count=count(t.y))

    A mapping of those names to new leaves to use in another compuation
    >>> two  # doctest: +SKIP
    {'x_sum': x_sum, 'y_count': y_count}

    A mapping of computations to do for each column
    >>> three   # doctest: +SKIP
    {'a': x_sum, 'b': (x_sum + y_count) - 1}

    In this way, ``compute_by`` is able to do simple pandas reductions using
    groups.agg(...) and then do columnwise arithmetic afterwards.
    """
    seen_names.clear()
    name_dict.clear()
    exprs = pipe(expr.values,
                 map(Expr._traverse),
                 concat,
                 filter(lambda x: isinstance(x, Reduction)),
                 set)
    one = summary(**dict((_name(expr), expr) for expr in exprs))

    two = dict((_name(expr), symbol(_name(expr), datashape.var * expr.dshape))
               for expr in exprs)

    d = dict((expr, two[_name(expr)]) for expr in exprs)
    three = dict((name, value._subs(d)) for name, value in zip(expr.names,
                                                               expr.values))

    return one, two, three


@dispatch(By, Summary, Grouper, NDFrame)
def compute_by(t, s, g, df):
    one, two, three = fancify_summary(s)  # see above

    names_columns = list(zip(one.fields, one.values))
    func = lambda x: not isinstance(x[1], count)
    is_field = defaultdict(lambda: iter([]), groupby(func, names_columns))

    preapply = DataFrame(dict(
        zip([name for name, _ in is_field[True]],
            [compute(col._child, {t._child: df}, return_type='native')
             for _, col in is_field[True]])
        )
    )

    if list(is_field[False]):
        emptys = DataFrame([0] * len(df.index),
                           index=df.index,
                           columns=[name for name, _ in is_field[False]])
        preapply = concat_nodup(preapply, emptys)
    if not df.index.equals(preapply.index):
        df = df.loc[preapply.index]
    df2 = concat_nodup(df, preapply)

    groups = df2.groupby(g)

    d = dict((name, v.symbol) for name, v in zip(one.names, one.values))

    result = groups.agg(d)

    scope = dict((v, result[k]) for k, v in two.items())
    cols = [
        compute(expr.label(name), scope, return_type='native')
        for name, expr in three.items()
    ]

    result2 = pd.concat(cols, axis=1)

    # Rearrange columns to match names order
    result3 = result2[
        sorted(result2.columns, key=lambda t, s=s: s.fields.index(t))
    ]
    return result3


@dispatch(Expr, (DataFrame, DaskDataFrame))
def post_compute_by(t, df):
    return df.reset_index(drop=True)


@dispatch((Summary, Reduction), (DataFrame, DaskDataFrame))
def post_compute_by(t, df):
    return df.reset_index()


@dispatch(By, NDFrame)
def compute_up(t, df, **kwargs):
    grouper = get_grouper(t, t.grouper, df)
    result = compute_by(t, t.apply, grouper, df)
    result2 = post_compute_by(t.apply, into(DataFrame, result))
    if isinstance(result2, DataFrame):
        result2.columns = t.fields
    return result2


def concat_nodup(a, b):
    """ Concatenate two dataframes/series without duplicately named columns


    >>> df = DataFrame([[1, 'Alice',   100],
    ...                 [2, 'Bob',    -200],
    ...                 [3, 'Charlie', 300]],
    ...                columns=['id','name', 'amount'])

    >>> concat_nodup(df, df)
       id     name  amount
    0   1    Alice     100
    1   2      Bob    -200
    2   3  Charlie     300


    >>> concat_nodup(df.name, df.amount)
          name  amount
    0    Alice     100
    1      Bob    -200
    2  Charlie     300



    >>> concat_nodup(df, df.amount + df.id)
       id     name  amount    0
    0   1    Alice     100  101
    1   2      Bob    -200 -198
    2   3  Charlie     300  303
    """

    if isinstance(a, DataFrame) and isinstance(b, DataFrame):
        return pd.concat([a, b[[c for c in b.columns if c not in a.columns]]],
                         axis=1)
    if isinstance(a, DataFrame) and isinstance(b, Series):
        if b.name not in a.columns:
            return pd.concat([a, b], axis=1)
        else:
            return a
    if isinstance(a, Series) and isinstance(b, DataFrame):
        return pd.concat([a, b[[c for c in b.columns if c != a.name]]], axis=1)
    if isinstance(a, Series) and isinstance(b, Series):
        if a.name == b.name:
            return a
        else:
            return pd.concat([a, b], axis=1)


pdsort = getattr(
    pd.DataFrame,
    'sort' if LooseVersion(pd.__version__) < '0.17.0' else 'sort_values'
)


@dispatch(Sort, DataFrame)
def compute_up(t, df, **kwargs):
    return pdsort(df, t.key, ascending=t.ascending)


@dispatch(Sort, Series)
def compute_up(t, s, **kwargs):
    try:
        return s.sort_values(ascending=t.ascending)
    except AttributeError:
        return s.order(ascending=t.ascending)


@dispatch(Sample, (Series, DataFrame))
def compute_up(t, df, **kwargs):
    from math import modf
    if t.frac is not None:
        # Work around annoying edge case: Python's round() builtin (which
        # Pandas' sample() uses) rounds 0.5, 2.5, 4.5, ... down to 0, 2, 4, ...,
        # while it rounds 1.5, 3.5, 5.5, ... up.  This is inconsistent with any
        # sane implementation of floating point rounding, including SQL's, so
        # we work around it here.
        fractional, integral = modf(t.frac * df.shape[0])
        n = int(integral + (0 if fractional < 0.5 else 1))
    else:
        n = min(t.n, df.shape[0])
    return df.sample(n=n)


@dispatch(Sample, (DaskDataFrame, DaskSeries))
def compute_up(t, df, **kwargs):
    # Dask doesn't support sample(n=...), only sample(frac=...), so we have a
    # separate dispatch for dask objects.
    if t.frac is not None:
        frac = t.frac
    else:
        nrows = len(df)
        frac = float(min(t.n, nrows)) / nrows
    return df.sample(frac=frac)


@dispatch(Head, (Series, DataFrame, DaskDataFrame, DaskSeries))
def compute_up(t, df, **kwargs):
    return df.head(t.n)


@dispatch(Tail, (Series, DataFrame, DaskDataFrame, DaskSeries))
def compute_up(t, df, **kwargs):
    return df.tail(t.n)


@dispatch(Label, DataFrame)
def compute_up(t, df, **kwargs):
    return type(df)(df, columns=[t.label])


@dispatch(Label, Series)
def compute_up(t, df, **kwargs):
    return Series(df, name=t.label)


@dispatch(ReLabel, (DataFrame, DaskDataFrame))
def compute_up(t, df, **kwargs):
    return df.rename(columns=dict(t.labels))


@dispatch(ReLabel, (Series, DaskSeries))
def compute_up(t, s, **kwargs):
    labels = t.labels
    if len(labels) > 1:
        raise ValueError('You can only relabel a Series with a single name')
    pair, = labels
    _, replacement = pair
    return Series(s, name=replacement)


@dispatch(Map, DataFrame)
def compute_up(t, df, **kwargs):
    return df.apply(lambda tup: t.func(*tup), axis=1)


@dispatch(Map, Series)
def compute_up(t, df, **kwargs):
    result = df.map(t.func)
    try:
        result.name = t._name
    except NotImplementedError:
        # We don't have a schema, but we should still be able to map
        result.name = df.name
    return result


@dispatch(Apply, (Series, DataFrame))
def compute_up(t, df, **kwargs):
    return t.func(df)


@dispatch(Merge, NDFrame)
def compute_up(t, df, scope=None, **kwargs):
    subexpression = common_subexpression(*t.children)
    scope = merge_dicts(scope or {}, {subexpression: df})
    children = [
        compute(_child, scope, return_type='native')
        for _child in t.children
    ]
    return pd.concat(children, axis=1)


@dispatch(Summary, (DataFrame, DaskDataFrame))
def compute_up(expr, data, **kwargs):
    values = [
        compute(val, {expr._child: data}, return_type='native')
        for val in expr.values
    ]
    if expr.keepdims:
        return type(data)([values], columns=expr.fields)
    else:
        return Series(dict(zip(expr.fields, values)))


@dispatch(Summary, (Series, DaskSeries))
def compute_up(expr, data, **kwargs):
    result = tuple(
        compute(val, {expr._child: data}, return_type='native')
        for val in expr.values
    )
    if expr.keepdims:
        result = [result]
    return result


@dispatch(Like, Series)
def compute_up(expr, data, **kwargs):
    return data.str.contains(r'^%s$' % fnmatch.translate(expr.pattern))


def get_date_attr(s, attr, name):
    try:
        result = getattr(s.dt, attr)  # new in pandas 0.15
    except AttributeError:
        result = getattr(pd.DatetimeIndex(s), attr)
    result.name = name
    return result


@dispatch(DateTime, Series)
def compute_up(expr, s, **kwargs):
    return get_date_attr(s, expr.attr, expr._name)


@dispatch(total_seconds, Series)
def compute_up(expr, s, **kwargs):
    result = s.dt.total_seconds()
    result.name = expr._name
    return result

@dispatch(UTCFromTimestamp, Series)
def compute_up(expr, s, **kwargs):
    return pd.datetools.to_datetime(s * 1e9, utc=True)


@dispatch(Millisecond, Series)
def compute_up(expr, s, **kwargs):
    return get_date_attr(s, 'microsecond',
                         '%s_millisecond' % expr._child._name) // 1000


@dispatch(Round, Series)
def compute_up(expr, data, **kwargs):
    return data.dt.round(expr.freq)


@dispatch(Ceil, Series)
def compute_up(expr, data, **kwargs):
    return data.dt.ceil(expr.freq)


@dispatch(Floor, Series)
def compute_up(expr, data, **kwargs):
    return data.dt.floor(expr.freq)


@dispatch(strftime, Series)
def compute_up(expr, data, **kwargs):
    return data.dt.strftime(expr.format)


@dispatch(Slice, (DataFrame, Series))
def compute_up(expr, df, **kwargs):
    index = expr.index
    if isinstance(index, tuple) and len(index) == 1:
        index = index[0]
    if isinstance(index, _inttypes + (list,)):
        return df.iloc[index]
    elif isinstance(index, slice):
        if index.stop is not None:
            return df.iloc[index.start:index.stop:index.step]
        else:
            return df.iloc[index]
    else:
        raise NotImplementedError()


@dispatch(count, DataFrame)
def compute_up(expr, df, **kwargs):
    result = df.shape[0]
    if expr.keepdims:
        result = Series([result], name=expr._name)
    return result


@dispatch(nelements, (DataFrame, Series))
def compute_up(expr, df, **kwargs):
    return df.shape[0]


@dispatch((count, nelements), (DaskDataFrame, DaskSeries))
def compute_up(expr, df, **kwargs):
    warnings.warn("Counting the elements of a dask object can be slow.")
    result = len(df)
    if expr.keepdims:
        result = DaskSeries([result], name=expr._name)
    return result


@dispatch(DateTimeTruncate, Series)
def compute_up(expr, data, **kwargs):
    return Series(compute_up(expr, into(np.ndarray, data), **kwargs),
                  name=expr._name)


@dispatch(IsIn, (Series, DaskSeries))
def compute_up(expr, data, **kwargs):
    return data.isin(expr._keys)


@dispatch(Coerce, (Series, DaskSeries))
def compute_up(expr, data, **kwargs):
    measure = expr.to.measure
    if measure in {datashape.string, datashape.Option(datashape.string)}:
        return data.astype(str)
    elif measure in {datashape.datetime_,
                     datashape.Option(datashape.datetime_)}:
        return data.astype(np.datetime64)
    return data.astype(to_numpy_dtype(expr.schema))


@dispatch(Shift, Series)
def compute_up(expr, data, **kwargs):
    return data.shift(expr.n)


def array_coalesce(expr, lhs, rhs, wrap=None, **kwargs):
    res = np.where(pd.isnull(lhs), rhs, lhs)
    if not expr.dshape.shape:
        res = res.item()
    elif wrap:
        res = wrap(res)
    return res


@compute_up.register(
    Coalesce, (Series, DaskSeries), (np.ndarray, Series, base, DaskSeries)
)
@compute_up.register(
    Coalesce,
    (Series, base, DaskSeries), (np.ndarray, Series, DaskSeries)
)
def compute_up_coalesce(expr, lhs, rhs, **kwargs):
    return array_coalesce(expr, lhs, rhs, type(lhs))



@dispatch(Coalesce, (Series, DaskSeries, base))
def compute_up(t, data, **kwargs):
    if isinstance(t.lhs, Expr):
        lhs = data
        rhs = t.rhs
    else:
        lhs = t.lhs
        rhs = data

    return compute_up_coalesce(t, lhs, rhs)



from __future__ import absolute_import, division, print_function

import pytest
import sys

dask = pytest.importorskip('dask')

import numpy as np

from dask.array import Array, from_array
from odo import discover
from operator import getitem

from blaze import compute, symbol


def eq(a, b):
    if isinstance(a, Array):
        a = a.compute()
    if isinstance(b, Array):
        b = b.compute()
    c = a == b
    if isinstance(c, np.ndarray):
        c = c.all()
    return c


nx = np.arange(600).reshape((20, 30))
dx = from_array(nx, (4, 5))
sx = symbol('x', discover(dx))

ny = np.arange(600).reshape((30, 20))
dy = from_array(ny, (5, 4))
sy = symbol('y', discover(dy))

na = np.arange(20)
da = from_array(na, (4,))
sa = symbol('a', discover(da))

nb = np.arange(30).reshape((30, 1))
db = from_array(nb, (5, 1))
sb = symbol('b', discover(db))

dask_ns = {sx: dx, sy: dy, sa: da, sb: db}
numpy_ns = {sx: nx, sy: ny, sa: na, sb: nb}

exprs = [2 * sx + 1,
         sx.sum(axis=0),
         pytest.mark.skipif(True, sx.mean(axis=0),
                            reason="Segfaulting and remote process error"),
         sx + sx,
         sx.T,
         sx.T + sy,
         sx.dot(sy),
         sy.dot(sx),
         sx.sum(),
         sx - sx.sum(),
         sx.dot(sx.T),
         sx.sum(axis=1),
         sy + sa,
         sy + sb,
         sx[3:17],
         sx[3:10, 10:25:2] + 1,
         sx[:5, 10],
         sx[0, 0]]


@pytest.mark.parametrize('expr', exprs)
@pytest.mark.skipif(sys.platform == 'win32' and sys.version_info[:2] == (2, 7),
		    reason="Nondeterministic exceptions relating to get_async")
def test_compute(expr):
    result = compute(expr, dask_ns)
    expected = compute(expr, numpy_ns)
    assert isinstance(result, Array)
    if expr.dshape.shape:
        result2 = np.array(result)
    else:
        result2 = float(result)
    assert eq(result2, expected)


def test_elemwise_broadcasting():
    arr = compute(sy + sb, dask_ns)
    expected = [[(arr.name, i, j) for j in range(5)]
                for i in range(6)]
    assert arr._keys() == expected


def test_ragged_blockdims():
    dsk = {('x', 0, 0): np.ones((2, 2)),
           ('x', 0, 1): np.ones((2, 3)),
           ('x', 1, 0): np.ones((5, 2)),
           ('x', 1, 1): np.ones((5, 3))}

    a = Array(dsk, 'x', chunks=[(2, 5), (2, 3)], shape=(7, 5))
    s = symbol('s', '7 * 5 * int')

    assert compute(s.sum(axis=0), a).chunks == ((2, 3),)
    assert compute(s.sum(axis=1), a).chunks == ((2, 5),)

    assert compute(s + 1, a).chunks == a.chunks


def test_slicing_with_singleton_dimensions():
    arr = compute(sx[5:15, 12], dx)
    x = dx.name
    y = arr.name
    assert arr.dask[(y, 0)] == (getitem, (x, 1, 2), (slice(1, 4, 1), 2))
    assert arr.dask[(y, 1)] == (
        getitem, (x, 2, 2), (slice(None, None, None), 2))
    assert arr.dask[(y, 2)] == (getitem, (x, 3, 2), (slice(0, 3, 1), 2))
    assert all(len(k) == 2 for k in arr._keys())


def test_slicing_with_lists():
    nx = np.arange(20).reshape((4, 5))
    dx = from_array(nx, (2, 2))
    sx = symbol('x', discover(dx))

    expr = sx[[2, 0, 3]]
    assert eq(np.array(compute(expr, dx)), compute(expr, nx))

    expr = sx[::2, [2, 0, 3]]
    assert eq(np.array(compute(expr, dx)), compute(expr, nx))

    expr = sx[1, [2, 0, 3]]
    assert eq(np.array(compute(expr, dx)), compute(expr, nx))

    expr = sx[[2, 0, 3], -2]
    assert eq(np.array(compute(expr, dx)), compute(expr, nx))

    expr = sx[:, :]
    assert compute(expr, dx).dask == dx.dask

    expr = sx[0]
    assert eq(np.array(compute(expr, dx)), compute(expr, nx))

    expr = sx[0, [3, 1, 4]]
    assert eq(np.array(compute(expr, dx)), compute(expr, nx))


def test_slicing_on_boundary_lines():
    nx = np.arange(100).reshape((10, 10))
    dx = from_array(nx, (3, 3))
    sx = symbol('x', discover(dx))
    expr = sx[0, [1, 3, 9, 3]]

    result = compute(expr, dx)
    assert eq(result, nx[0, [1, 3, 9, 3]])


def test_slicing_with_newaxis():
    nx = np.arange(20).reshape((4, 5))
    dx = from_array(nx, (2, 2))
    sx = symbol('x', discover(dx))

    expr = sx[:, None, :]
    result = compute(expr, dx)

    assert result.shape == (4, 1, 5)
    assert result.chunks == ((2, 2), (1,), (2, 2, 1))
    assert eq(np.array(result), compute(expr, nx))

    expr = sx[None, [2, 1, 3], None, None, :, None]
    result = compute(expr, dx)

    assert result.shape == (1, 3, 1, 1, 5, 1)
    assert result.chunks == ((1,), (3,), (1,), (1,), (2, 2, 1), (1,))
    assert eq(np.array(result), compute(expr, nx))



#!/usr/bin/env python
from __future__ import absolute_import

import os
import argparse
import importlib

from contextlib import contextmanager

import yaml

from blaze.interactive import data as bz_data
from odo.utils import ignoring

from .server import Server, DEFAULT_PORT

try:
    import __builtin__ as builtins
except ImportError:
    import builtins


__all__ = 'data_spider', 'from_yaml'


def _spider(resource_path, ignore, followlinks, hidden, extra_kwargs):
    resources = {}
    for filename in (os.path.join(resource_path, x)
                     for x in os.listdir(resource_path)):
        basename = os.path.basename(filename)
        if (basename.startswith(os.curdir) and not hidden or
                os.path.islink(filename) and not followlinks):
            continue
        if os.path.isdir(filename):
            new_resources = _spider(filename, ignore=ignore,
                                    followlinks=followlinks,
                                    hidden=hidden,
                                    extra_kwargs=extra_kwargs)
            if new_resources:
                resources[basename] = new_resources
        else:
            with ignoring(*ignore):
                resources[basename] = bz_data(filename, **(extra_kwargs or {}))
    return resources


def data_spider(path,
                ignore=(ValueError, NotImplementedError),
                followlinks=True,
                hidden=False,
                extra_kwargs=None):
    """Traverse a directory and call ``blaze.data`` on its contents.

    Parameters
    ----------
    path : str
        Path to a directory of resources to load
    ignore : tuple of Exception, optional
        Ignore these exceptions when calling ``blaze.data``
    followlinks : bool, optional
        Follow symbolic links
    hidden : bool, optional
        Load hidden files
    extra_kwargs: dict, optional
        extra kwargs to forward on to ``blaze.data``

    Returns
    -------
    dict
        Possibly nested dictionary of containing basenames mapping to resources
    """
    # NOTE: this is named `data_spider` rather than just `spider` to
    # disambiguate this function from the `blaze.server.spider` module.
    return {os.path.basename(path): _spider(path,
                                            ignore=ignore,
                                            followlinks=followlinks,
                                            hidden=hidden,
                                            extra_kwargs=extra_kwargs)}


@contextmanager
def pushd(path):
    """Context manager that changes to ``path`` directory on enter and
    changes back to ``os.getcwd()`` on exit.
    """
    cwd = os.getcwd()
    os.chdir(os.path.abspath(path))
    try:
        yield
    finally:
        os.chdir(cwd)


def from_yaml(fh,
              ignore=(ValueError, NotImplementedError),
              followlinks=True,
              hidden=False,
              relative_to_yaml_dir=False):
    """Construct a dictionary of resources from a YAML specification.

    Parameters
    ----------
    fh : file
        File object referring to the YAML specification of resources to load.
    ignore : tuple of Exception, optional
        Ignore these exceptions when calling ``blaze.data``.
    followlinks : bool, optional
        Follow symbolic links.
    hidden : bool, optional
        Load hidden files.
    relative_to_yaml_dir: bool, optional, default False
        Load paths relative to yaml file's directory.  Default is to load
        relative to process' CWD.

    Returns
    -------
    dict
        A dictionary mapping top level keys in a YAML file to resources.

    See Also
    --------
    data_spider : Traverse a directory tree for resources
    """
    resources = {}
    yaml_dir = os.path.dirname(os.path.abspath(fh.name))
    for name, info in yaml.load(fh.read()).items():
        with pushd(yaml_dir if relative_to_yaml_dir else os.getcwd()):
            try:
                source = info.pop('source')
            except KeyError:
                raise ValueError('source key not found for data source named %r' %
                                 name)
            for mod in info.pop('imports', []):
                importlib.import_module(mod)
            if os.path.isdir(source):
                resources[name] = data_spider(os.path.expanduser(source),
                                              ignore=ignore,
                                              followlinks=followlinks,
                                              hidden=hidden,
                                              extra_kwargs=info)
            else:
                resources[name] = bz_data(source, **info)
    return resources


def _parse_args():
    p = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    p.add_argument('path', type=argparse.FileType('r'), nargs='?',
                   default=None,
                   help='A YAML file specifying the resources to load')
    p.add_argument('-p', '--port', type=int, default=DEFAULT_PORT,
                   help='Port number')
    p.add_argument('-H', '--host', type=str, default='127.0.0.1',
                   help='Host name. Use 0.0.0.0 to listen on all public IPs')
    p.add_argument('-l', '--follow-links', action='store_true',
                   help='Follow links when listing files')
    p.add_argument('-e', '--ignored-exception', nargs='+',
                   default=['Exception'],
                   help='Exceptions to ignore when calling ``blaze.data`` on a file')
    p.add_argument('-d', '--hidden', action='store_true',
                   help='Call ``blaze.data`` on hidden files')
    p.add_argument('--yaml-dir', action='store_true',
                   help='Load path-based resources relative to yaml file directory.')
    p.add_argument('--allow-dynamic-addition', action='store_true',
                   help='Allow dynamically adding datasets to the server')
    p.add_argument('-D', '--debug', action='store_true',
                   help='Start the Flask server in debug mode')
    p.add_argument('--log-file', type=str, default=None,
                   help='Log filename -- if not set, logs to standard output.')
    p.add_argument('--log-level', type=str, default='WARNING',
                   help=('Level of output for logs. Set to DEBUG for most'
                         'verbose, or ERROR for least.'))
    args = p.parse_args()
    if not (args.path or args.allow_dynamic_addition):
        msg = "No YAML file provided and --allow-dynamic-addition flag not set."
        p.error(msg)
    return args


def _main():
    args = _parse_args()
    ignore = tuple(getattr(builtins, e) for e in args.ignored_exception)
    if args.path:
        resources = from_yaml(args.path,
                              ignore=ignore,
                              followlinks=args.follow_links,
                              hidden=args.hidden,
                              relative_to_yaml_dir=args.yaml_dir)
    else:
        resources = {}
    server = Server(resources,
                    allow_add=args.allow_dynamic_addition,
                    logfile=args.log_file,
                    loglevel=args.log_level)
    server.run(host=args.host, port=args.port, debug=args.debug)


if __name__ == '__main__':
    _main()



from blaze.cached import CachedDataset
from blaze import symbol, discover, compute, into
import pandas as pd
from collections import Iterator


df = pd.DataFrame([['Alice', 100, 1],
                   ['Bob', 200, 2],
                   ['Alice', 50, 3]],
                  columns=['name', 'amount', 'id'])


t = symbol('t', discover(df))


def test_dataset():
    ns = {'t': df, 'x': 10}
    cache=dict()
    d = CachedDataset(ns, cache=cache)

    assert discover(d) == discover(ns)

    s = symbol('s', discover(d))
    compute(s.x * 2, d) == 20
    cache == {s.x * 2: 20}


def test_streaming():
    seq = [{'name': 'Alice', 'x': 1},
           {'name': 'Bob', 'x': 1}]
    ns = {'t': seq, 'x': 10}
    cache=dict()
    d = CachedDataset(ns, cache=cache)

    s = symbol('s', discover(d))
    expr = s.t.x * 2
    result = compute(expr, d)

    assert not isinstance(d.cache[expr], Iterator)
    assert into(list, d.cache[expr]) == [2, 2]



from datetime import timedelta
from operator import methodcaller
import itertools
import math

import pytest

sa = pytest.importorskip('sqlalchemy')
pytest.importorskip('psycopg2')

import os
import numpy as np
import pandas as pd

import pandas.util.testing as tm

from datashape import dshape
from odo import odo, drop, discover
from blaze import (
    data,
    atan2,
    by,
    coalesce,
    compute,
    concat,
    cos,
    greatest,
    join,
    least,
    radians,
    sin,
    sqrt,
    symbol,
    transform,
)
from blaze.interactive import iscorescalar
from blaze.utils import example, normalize


names = ('tbl%d' % i for i in itertools.count())

@pytest.fixture(scope='module')
def pg_ip():
    return os.environ.get('POSTGRES_IP', 'localhost')

@pytest.fixture
def url(pg_ip):
    return 'postgresql://postgres@{}/test::%s'.format(pg_ip)


@pytest.yield_fixture
def sql(url):
    ds = dshape('var * {A: string, B: int64}')
    try:
        t = data(url % next(names), dshape=ds)
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        assert t.dshape == ds
        t = data(odo([('a', 1), ('b', 2)], t))
        try:
            yield t
        finally:
            drop(t)

@pytest.yield_fixture
def sql_with_null(url):
    ds = dshape(""" var * {name: ?string,
                           sex: ?string,
                           amount: int,
                           id: int,
                           comment: ?string}
              """)
    rows = [('Alice', 'F', 100, 1, 'Alice comment'),
            (None, 'M', 300, 2, None),
            ('Drew', 'F', 100, 4, 'Drew comment'),
            ('Bob', 'M', 100, 5, 'Bob comment 2'),
            ('Drew', 'M', 200, 5, None),
            ('first', None, 300, 4, 'Missing info'),
            (None, None, 300, 6, None)]
    try:
        x = url % next(names)
        t = data(x, dshape=ds)
        print(x)
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        assert t.dshape == ds
        t = data(odo(rows, t))
        try:
            yield t
        finally:
            drop(t)


@pytest.yield_fixture(scope='module')
def nyc(pg_ip):
    # odoing csv -> pandas -> postgres is more robust, as it doesn't require
    # the postgres server to be on the same filesystem as the csv file.
    nyc_pd = odo(example('nyc.csv'), pd.DataFrame)
    try:
        t = odo(nyc_pd,
                'postgresql://postgres@{}/test::nyc'.format(pg_ip))
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            yield t
        finally:
            drop(t)


@pytest.yield_fixture
def big_sql(url):
    try:
        t = data(url % next(names), dshape='var * {A: string, B: int64}')
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        t = odo(zip(list('a'*100), list(range(100))), t)
        try:
            yield t
        finally:
            drop(t)


@pytest.yield_fixture
def sqla(url):
    try:
        t = data(url % next(names), dshape='var * {A: ?string, B: ?int32}')
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        t = odo([('a', 1), (None, 1), ('c', None)], t)
        try:
            yield t
        finally:
            drop(t)


@pytest.yield_fixture
def sqlb(url):
    try:
        t = data(url % next(names), dshape='var * {A: string, B: int64}')
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        t = odo([('a', 1), ('b', 2)], t)
        try:
            yield t
        finally:
            drop(t)


@pytest.yield_fixture
def sql_with_dts(url):
    try:
        t = data(url % next(names), dshape='var * {A: datetime}')
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        t = odo([(d,) for d in pd.date_range('2014-01-01', '2014-02-01')], t)
        try:
            yield t
        finally:
            drop(t)


@pytest.yield_fixture
def sql_with_timedeltas(url):
    try:
        t = data(url % next(names), dshape='var * {N: timedelta}')
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        t = odo([(timedelta(seconds=n),) for n in range(10)], t)
        try:
            yield t
        finally:
            drop(t)


@pytest.yield_fixture
def sql_two_tables(url):
    dshape = 'var * {a: int32}'
    try:
        t = data(url % next(names), dshape=dshape)
        u = data(url % next(names), dshape=dshape)
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            yield u, t
        finally:
            drop(t)
            drop(u)


@pytest.yield_fixture
def products(url):
    try:
        products = data(url % 'products',
                        dshape="""var * {
                            product_id: int64,
                            color: ?string,
                            price: float64}""",
                        primary_key=['product_id'])
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            yield products
        finally:
            drop(products)


@pytest.yield_fixture
def orders(url, products):
    try:
        orders = data(url % 'orders',
                      dshape="""var * {
                        order_id: int64,
                        product_id: map[int64,
                                        {product_id: int64,
                                         color: ?string,
                                         price: float64}],
                        quantity: int64}""",
                      foreign_keys=dict(product_id=products.data.c.product_id),
                      primary_key=['order_id'])
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            yield orders
        finally:
            drop(orders)


# TODO: scope these as module because I think pytest is caching sa.Table, which
# doesn't work if remove it after every run

@pytest.yield_fixture
def main(url):
    try:
        main = odo([(i, int(np.random.randint(10))) for i in range(13)],
                   url % 'main',
                   dshape=dshape('var * {id: int64, data: int64}'),
                   primary_key=['id'])
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            yield main
        finally:
            drop(main)


@pytest.yield_fixture
def pkey(url, main):
    choices = [u'AAPL', u'HPQ', u'ORCL', u'IBM', u'DOW', u'SBUX', u'AMD',
               u'INTC', u'GOOG', u'PRU', u'MSFT', u'AIG', u'TXN', u'DELL',
               u'PEP']
    n = 100
    data = list(zip(range(n),
                    np.random.choice(choices, size=n).tolist(),
                    np.random.uniform(10000, 20000, size=n).tolist(),
                    np.random.randint(main.count().scalar(), size=n).tolist()))
    try:
        pkey = odo(data, url % 'pkey',
                dshape=dshape('''var * {id: int64,
                                        sym: string,
                                        price: float64,
                                        main: map[int64,
                                                  {id: int64,
                                                   data: int64}]}'''),
                   foreign_keys=dict(main=main.c.id),
                   primary_key=['id'])
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            yield pkey
        finally:
            drop(pkey)


@pytest.yield_fixture
def fkey(url, pkey):
    try:
        fkey = odo([(i,
                     int(np.random.randint(pkey.count().scalar())),
                     int(np.random.randint(10000)))
                    for i in range(10)],
                   url % 'fkey',
                   dshape=dshape('''var * {id: int64,
                                           sym_id: map[int64,
                                                       {id: int64,
                                                        sym: string,
                                                        price: float64,
                                                        main: map[int64,
                                                                  {id: int64,
                                                                  data: int64}]}],
                                           size: int64}'''),
                   foreign_keys=dict(sym_id=pkey.c.id),
                   primary_key=['id'])
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            yield fkey
        finally:
            drop(fkey)


@pytest.yield_fixture
def sql_with_float(url):
    try:
        t = data(url % next(names), dshape='var * {c: float64}')
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            yield t
        finally:
            drop(t)


@pytest.yield_fixture(scope='module')
def nyc_csv(pg_ip):
    try:
        t = odo(
            example('nyc.csv'),
            'postgresql://postgres@{}/test::nyc'.format(pg_ip),
        )
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            yield t
        finally:
            drop(t)


def test_nyc_csv(nyc_csv):
    t = symbol('t', discover(nyc_csv))
    assert compute(t.nrows, nyc_csv, return_type='core') > 0


def test_postgres_create(sql):
    assert odo(sql, list) == [('a', 1), ('b', 2)]


def test_postgres_isnan(sql_with_float):
    dta = (1.0,), (float('nan'),)
    table = odo(dta, sql_with_float)
    sym = symbol('s', discover(dta))
    assert compute(sym.isnan(), table, return_type=list) == [False, True]


def test_insert_from_subselect(sql_with_float):
    data = pd.DataFrame([{'c': 2.0}, {'c': 1.0}])
    tbl = odo(data, sql_with_float)
    s = symbol('s', discover(data))
    odo(compute(s[s.c.isin((1.0, 2.0))].sort(), tbl, return_type='native'), sql_with_float),
    tm.assert_frame_equal(
        odo(sql_with_float, pd.DataFrame).iloc[2:].reset_index(drop=True),
        pd.DataFrame([{'c': 1.0}, {'c': 2.0}]),
    )


def test_concat(sql_two_tables):
    t_table, u_table = sql_two_tables
    t_data = pd.DataFrame(np.arange(5), columns=['a'])
    u_data = pd.DataFrame(np.arange(5, 10), columns=['a'])
    odo(t_data, t_table)
    odo(u_data, u_table)

    t = symbol('t', discover(t_data))
    u = symbol('u', discover(u_data))
    tm.assert_frame_equal(
        compute(concat(t, u).sort('a'), {t: t_table, u: u_table}, return_type=pd.DataFrame),
        pd.DataFrame(np.arange(10), columns=['a']),
    )


def test_concat_invalid_axis(sql_two_tables):
    t_table, u_table = sql_two_tables
    t_data = pd.DataFrame(np.arange(5), columns=['a'])
    u_data = pd.DataFrame(np.arange(5, 10), columns=['a'])
    odo(t_data, t_table)
    odo(u_data, u_table)

    # We need to force the shape to not be a record here so we can
    # create the `Concat` node with an axis=1.
    t = symbol('t', '5 * 1 * int32')
    u = symbol('u', '5 * 1 * int32')

    with pytest.raises(ValueError) as e:
        compute(concat(t, u, axis=1), {t: t_table, u: u_table}, return_type='native')

    # Preserve the suggestion to use merge.
    assert "'merge'" in str(e.value)


def test_timedelta_arith(sql_with_dts):
    delta = timedelta(days=1)
    dates = pd.Series(pd.date_range('2014-01-01', '2014-02-01'))
    sym = symbol('s', discover(dates))
    assert (
        compute(sym + delta, sql_with_dts, return_type=pd.Series) == dates + delta
    ).all()
    assert (
        compute(sym - delta, sql_with_dts, return_type=pd.Series) == dates - delta
    ).all()
    assert (
        compute(sym - (sym - delta), sql_with_dts, return_type=pd.Series) ==
        (dates - (dates - delta))
    ).all()


@pytest.mark.parametrize('func', ('var', 'std'))
def test_timedelta_stat_reduction(sql_with_timedeltas, func):
    sym = symbol('s', discover(sql_with_timedeltas))
    expr = getattr(sym.N, func)()

    deltas = pd.Series([timedelta(seconds=n) for n in range(10)])
    expected = timedelta(
        seconds=getattr(deltas.astype('int64') / 1e9, func)(ddof=expr.unbiased)
    )
    assert compute(expr, sql_with_timedeltas, return_type=timedelta) == expected


def test_coerce_bool_and_sum(sql):
    sql = sql.data
    n = sql.name
    t = symbol(n, discover(sql))
    expr = (t.B > 1.0).coerce(to='int32').sum()
    result = compute(expr, sql).scalar()
    expected = compute(t.B, sql, return_type=pd.Series).gt(1).sum()
    assert result == expected


def test_distinct_on(sql):
    sql = sql.data
    t = symbol('t', discover(sql))
    computation = compute(t[['A', 'B']].sort('A').distinct('A'), sql, return_type='native')
    assert normalize(str(computation)) == normalize("""
    SELECT DISTINCT ON (anon_1."A") anon_1."A", anon_1."B"
    FROM (SELECT {tbl}."A" AS "A", {tbl}."B" AS "B"
    FROM {tbl}) AS anon_1 ORDER BY anon_1."A" ASC
    """.format(tbl=sql.name))
    assert odo(computation, tuple) == (('a', 1), ('b', 2))


def test_relabel_columns_over_selection(big_sql):
    t = symbol('t', discover(big_sql))
    result = compute(t[t['B'] == 2].relabel(B=u'b'),
                     big_sql, return_type=pd.DataFrame)
    expected = pd.DataFrame([['a', 2]], columns=[u'A', u'b'])
    tm.assert_frame_equal(result, expected)


def test_auto_join_field(orders):
    t = symbol('t', discover(orders))
    expr = t.product_id.color
    result = compute(expr, orders, return_type='native')
    expected = """SELECT
        products.color
    FROM products, orders
    WHERE orders.product_id = products.product_id
    """
    assert normalize(str(result)) == normalize(expected)


def test_auto_join_projection(orders):
    t = symbol('t', discover(orders))
    expr = t.product_id[['color', 'price']]
    result = compute(expr, orders, return_type='native')
    expected = """SELECT
        products.color,
        products.price
    FROM products, orders
    WHERE orders.product_id = products.product_id
    """
    assert normalize(str(result)) == normalize(expected)


@pytest.mark.xfail
@pytest.mark.parametrize('func', ['max', 'min', 'sum'])
def test_foreign_key_reduction(orders, products, func):
    t = symbol('t', discover(orders))
    expr = methodcaller(func)(t.product_id.price)
    result = compute(expr, orders, return_type='native')
    expected = """WITH alias as (select
            products.price as price
        from
            products, orders
        where orders.product_id = products.product_id)
    select {0}(alias.price) as price_{0} from alias
    """.format(func)
    assert normalize(str(result)) == normalize(expected)


def test_foreign_key_chain(fkey):
    t = symbol('t', discover(fkey))
    expr = t.sym_id.main.data
    result = compute(expr, fkey, return_type='native')
    expected = """SELECT
        main.data
    FROM main, fkey, pkey
    WHERE fkey.sym_id = pkey.id and pkey.main = main.id
    """
    assert normalize(str(result)) == normalize(expected)


@pytest.mark.xfail(raises=AssertionError,
                   reason='CTE mucks up generation here')
@pytest.mark.parametrize('grouper', ['sym', ['sym']])
def test_foreign_key_group_by(fkey, grouper):
    t = symbol('fkey', discover(fkey))
    expr = by(t.sym_id[grouper], avg_price=t.sym_id.price.mean())
    result = compute(expr, fkey, return_type='native')
    expected = """SELECT
        pkey.sym,
        avg(pkey.price) AS avg_price
    FROM pkey, fkey
    WHERE fkey.sym_id = pkey.id
    GROUP BY pkey.sym
    """
    assert normalize(str(result)) == normalize(expected)


@pytest.mark.parametrize('grouper', ['sym_id', ['sym_id']])
def test_group_by_map(fkey, grouper):
    t = symbol('fkey', discover(fkey))
    expr = by(t[grouper], id_count=t.size.count())
    result = compute(expr, fkey, return_type='native')
    expected = """SELECT
        fkey.sym_id,
        count(fkey.size) AS id_count
    FROM fkey
    GROUP BY fkey.sym_id
    """
    assert normalize(str(result)) == normalize(expected)


def test_foreign_key_isin(fkey):
    t = symbol('fkey', discover(fkey))
    expr = t.sym_id.isin([1, 2])
    result = compute(expr, fkey, return_type='native')
    expected = """SELECT
        fkey.sym_id IN (%(sym_id_1)s, %(sym_id_2)s) AS anon_1
    FROM fkey
    """
    assert normalize(str(result)) == normalize(expected)


@pytest.mark.xfail(raises=AssertionError, reason='Not yet implemented')
def test_foreign_key_merge_expression(fkey):
    from blaze import merge

    t = symbol('fkey', discover(fkey))
    expr = merge(t.sym_id.sym, t.sym_id.main.data)
    expected = """
        select pkey.sym, main.data
        from
            fkey, pkey, main
        where
            fkey.sym_id = pkey.id and pkey.main = main.id
    """
    result = compute(expr, fkey, return_type='native')
    assert normalize(str(result)) == normalize(expected)


def test_join_type_promotion(sqla, sqlb):
    t, s = symbol(sqla.name, discover(sqla)), symbol(sqlb.name, discover(sqlb))
    expr = join(t, s, 'B', how='inner')
    result = set(map(tuple, compute(expr, {t: sqla, s: sqlb}, return_type='native').execute().fetchall()))
    expected = set([(1, 'a', 'a'), (1, None, 'a')])
    assert result == expected


@pytest.mark.parametrize(['n', 'column'],
                         [(1, 'A'), (-1, 'A'),
                          (1, 'B'), (-1, 'B'),
                          (0, 'A'), (0, 'B')])
def test_shift_on_column(n, column, sql):
    sql = sql.data
    t = symbol('t', discover(sql))
    expr = t[column].shift(n)
    result = compute(expr, sql, return_type=pd.Series)
    expected = odo(sql, pd.DataFrame)[column].shift(n)
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize('n', [-1, 0, 1])
def test_shift_arithmetic(sql, n):
    t = symbol('t', discover(sql))
    expr = t.B - t.B.shift(n)
    result = compute(expr, sql, return_type=pd.Series)
    df = odo(sql, pd.DataFrame)
    expected = df.B - df.B.shift(n)
    tm.assert_series_equal(result, expected)


def test_dist(nyc):
    def distance(lat1, lon1, lat2, lon2, R=3959):
        # http://andrew.hedges.name/experiments/haversine/
        dlon = radians(lon2 - lon1)
        dlat = radians(lat2 - lat1)
        a = sin(dlat / 2.0) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2.0) ** 2
        return R * 2 * atan2(sqrt(a), sqrt(1 - a))

    t = symbol('t', discover(nyc))

    filtered = t[
        (t.pickup_latitude >= 40.477399) &
        (t.pickup_latitude <= 40.917577) &
        (t.dropoff_latitude >= 40.477399) &
        (t.dropoff_latitude <= 40.917577) &
        (t.pickup_longitude >= -74.259090) &
        (t.pickup_longitude <= -73.700272) &
        (t.dropoff_longitude >= -74.259090) &
        (t.dropoff_longitude <= -73.700272) &
        (t.passenger_count < 6)
    ]
    dist = distance(filtered.pickup_latitude, filtered.pickup_longitude,
                    filtered.dropoff_latitude, filtered.dropoff_longitude)
    transformed = transform(filtered, dist=dist)
    assert (
        compute(transformed.dist.max(), nyc, return_type=float) ==
        compute(transformed.dist, nyc, return_type=pd.Series).max()
    )


def test_multiple_columns_in_transform(nyc):
    t = symbol('t', discover(nyc))
    t = t[
        (t.pickup_latitude >= 40.477399) &
        (t.pickup_latitude <= 40.917577) &
        (t.dropoff_latitude >= 40.477399) &
        (t.dropoff_latitude <= 40.917577) &
        (t.pickup_longitude >= -74.259090) &
        (t.pickup_longitude <= -73.700272) &
        (t.dropoff_longitude >= -74.259090) &
        (t.dropoff_longitude <= -73.700272) &
        (t.passenger_count < 6)
    ]
    hours = t.trip_time_in_secs.coerce('float64') / 3600.0
    avg_speed_in_mph = t.trip_distance / hours
    d = transform(t, avg_speed_in_mph=avg_speed_in_mph, mycol=avg_speed_in_mph + 1)
    df = compute(d[d.avg_speed_in_mph <= 200], nyc, return_type=pd.DataFrame)
    assert not df.empty


def test_coerce_on_select(nyc):
    t = symbol('t', discover(nyc))
    t = t[
        (t.pickup_latitude >= 40.477399) &
        (t.pickup_latitude <= 40.917577) &
        (t.dropoff_latitude >= 40.477399) &
        (t.dropoff_latitude <= 40.917577) &
        (t.pickup_longitude >= -74.259090) &
        (t.pickup_longitude <= -73.700272) &
        (t.dropoff_longitude >= -74.259090) &
        (t.dropoff_longitude <= -73.700272) &
        (t.passenger_count < 6)
    ]
    t = transform(t, pass_count=t.passenger_count + 1)
    result = compute(t.pass_count.coerce('float64'), nyc, return_type='native')
    s = odo(result, pd.Series)
    expected = compute(t, nyc, return_type=pd.DataFrame) \
                      .passenger_count.astype('float64') + 1.0
    assert list(s) == list(expected)


def test_interactive_len(sql):
    t = data(sql)
    assert len(t) == int(t.count())


def test_sample_n(nyc):
    t = symbol('t', discover(nyc))
    result = compute(t.sample(n=14), nyc, return_type=pd.DataFrame)
    assert len(result) == 14


def test_sample_bounded(nyc):
    t = symbol('t', discover(nyc))
    nrows = compute(t.nrows, nyc, return_type=int)
    result = compute(t.sample(n=2*nrows), nyc, return_type=pd.DataFrame)
    assert len(result) == nrows


def test_sample_frac(nyc):
    t = symbol('t', discover(nyc))
    result = compute(t.sample(frac=0.5), nyc, return_type=pd.DataFrame)
    num_rows = compute(t.nrows, nyc, return_type=int)
    # *Sigh* have to do proper rounding manually; Python's round() builtin is
    # borked.
    fractional, integral = math.modf(num_rows * 0.5)
    assert int(integral + (0 if fractional < 0.5 else 1)) == len(result)


def test_sample(big_sql):
    nn = symbol('nn', discover(big_sql))
    nrows = odo(compute(nn.nrows, big_sql), int)
    result = compute(nn.sample(n=nrows // 2), big_sql, return_type=pd.DataFrame)
    assert len(result) == nrows // 2
    result2 = compute(nn.sample(frac=0.5), big_sql, return_type=pd.DataFrame)
    assert len(result) == len(result2)


@pytest.mark.parametrize('slc', [0, 1000,
                                 slice(0, 1), slice(None, 3),
                                 slice(0, None), slice(2, None)])
def test_str_slice(slc, sql_with_null):
    name_series = pd.Series(['Alice', None, 'Drew', 'Bob', 'Drew', 'first', None],       
                            name='substring_1')
    t = symbol('t', discover(sql_with_null))
    result = compute(t.name.str[slc], sql_with_null, return_type=pd.Series).fillna('zzz')
    result[result == ''] = 'zzz'
    expected = name_series.str[slc].fillna('zzz')
    tm.assert_series_equal(result, expected)


def test_str_find(sql_with_null):
    t = symbol('t', discover(sql_with_null))
    result = compute(t.name.str.find('e'), sql_with_null, return_type=pd.Series)
    expected = pd.Series([5.0, np.nan, 3.0, 0.0, 3.0, 0.0, np.nan], name='position_1')
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize("sep", [None, " -- "])
def test_str_cat_with_null(sql_with_null, sep):
    t = symbol('t', discover(sql_with_null))
    res = compute(t.name.str_cat(t.sex, sep=sep), sql_with_null,
                  return_type=list)
    cols = compute(t[['name', 'sex']], sql_with_null, return_type=list)

    for r, (n, s) in zip(res, cols):
        if n is None or s is None:
            assert r is None
        else:
            assert (r == n + s if sep is None else r == n + sep + s)


def test_chain_str_cat_with_null(sql_with_null):
    t = symbol('t', discover(sql_with_null))
    expr = (t.name
            .str_cat(t.comment, sep=' ++ ')
            .str_cat(t.sex, sep=' -- '))
    res = compute(expr, sql_with_null, return_type=list)
    cols = compute(t[['name', 'comment', 'sex']], sql_with_null,
                   return_type=list)

    for r, (n, c, s) in zip(res, cols):
        if n is None or c is None or s is None:
            assert r is None
        else:
            assert (r == n + ' ++ ' + c + ' -- ' + s)


def test_str_cat_bcast(sql_with_null):
    t = symbol('t', discover(sql_with_null))
    lit_sym = symbol('s', 'string')
    s = t[t.amount <= 200]
    result = compute(s.comment.str_cat(lit_sym, sep=' '),
                     {t: sql_with_null, lit_sym: '!!'},
                     return_type=pd.Series)
    df = compute(s, sql_with_null,
                 return_type=pd.DataFrame)
    expected = df.comment.str.cat(['!!']*len(df.comment), sep=' ')

    assert all(expected[~expected.isnull()] == result[~result.isnull()])
    assert all(expected[expected.isnull()].index == result[result.isnull()].index)

    

def test_str_cat_where_clause(sql_with_null):
    """
    Invokes the (Select, Select) path for compute_up
    """
    t = symbol('t', discover(sql_with_null))
    s = t[t.amount <= 200]
    c1 = s.comment.str_cat(s.sex, sep=' -- ')

    bres = compute(c1, sql_with_null, return_type=pd.Series)
    df_s = compute(s, sql_with_null, return_type=pd.DataFrame)
    exp = df_s.comment.str.cat(df_s.sex, ' -- ')

    assert all(exp[~exp.isnull()] == bres[~bres.isnull()])
    assert all(exp[exp.isnull()].index == bres[bres.isnull()].index)


def test_core_compute(nyc):
    t = symbol('t', discover(nyc))
    assert isinstance(compute(t, nyc, return_type='core'), pd.DataFrame)
    assert isinstance(compute(t.passenger_count, nyc, return_type='core'), pd.Series)
    assert iscorescalar(compute(t.passenger_count.mean(), nyc, return_type='core'))
    assert isinstance(compute(t, nyc, return_type=list), list)


@pytest.fixture
def gl_data(sql_two_tables):
    u_data, t_data = sql_two_tables
    # populate the tables with some data and return it
    return data(odo([(1,)], u_data)), data(odo([(2,)], t_data))


def test_greatest(gl_data):
    u, t = gl_data
    assert odo(greatest(u.a.max(), t.a.max()), int) == 2


def test_least(gl_data):
    u, t = gl_data
    assert odo(least(u.a.max(), t.a.max()), int) == 1


def test_coalesce(sqla):
    t = symbol('t', discover(sqla))
    assert (
        compute(coalesce(t.B, -1), {t: sqla}, return_type=list) ==
        [1, 1, -1]
    )
    assert (
        compute(coalesce(t.A, 'z'), {t: sqla}, return_type=list) ==
        ['a', 'z', 'c']
    )


def test_any(sql):
    s = symbol('s', discover(sql))
    assert compute((s.B == 1).any(), {s: sql}, return_type='core')
    assert compute((s.B == 2).any(), {s: sql}, return_type='core')
    assert compute(~(s.B == 3).any(), {s: sql}, return_type='core')


def test_all(sql):
    s = symbol('s', discover(sql))
    assert compute(s.B.isin({1, 2}).all(), {s: sql}, return_type='core')
    assert compute(~(s.B == 1).all(), {s: sql}, return_type='core')
    assert compute(~(s.B == 2).all(), {s: sql}, return_type='core')
    assert compute(~(s.B == 3).all(), {s: sql}, return_type='core')


def test_isin_selectable(sql):
    s = symbol('s', discover(sql))

    # wrap the resource in a select
    assert compute(s.B.isin({1, 3}),
                   sa.select(sql._resources()[sql].columns),
                   return_type=list) == [True, False]


def test_selection_selectable(sql):
    s = symbol('s', discover(sql))

    # wrap the resource in a select
    assert (compute(s[s.B.isin({1, 3})],
                    sa.select(sql._resources()[sql].columns),
                    return_type=pd.DataFrame) ==
            pd.DataFrame([['a', 1]],
                         columns=s.dshape.measure.names)).all().all()



from __future__ import absolute_import, division, print_function

from collections import defaultdict, Iterator, Mapping
from datetime import date, datetime, timedelta
import itertools
import numbers
import warnings

import toolz
from toolz import first, unique, assoc
from toolz.utils import no_default
import pandas as pd
from odo import odo

from ..compatibility import basestring
from ..expr import Expr, Field, Symbol, symbol, Join, Cast
from ..dispatch import dispatch
from ..interactive import coerce_core, into


__all__ = ['compute', 'compute_up']

base = numbers.Number, basestring, date, datetime, timedelta, type(None)


@dispatch(Expr, object)
def pre_compute(leaf, data, scope=None, **kwargs):
    """ Transform data prior to calling ``compute`` """
    return data


@dispatch(Expr, object)
def post_compute(expr, result, scope=None):
    """ Effects after the computation is complete """
    return result


@dispatch(Expr, object)
def optimize(expr, data):
    """ Optimize expression to be computed on data """
    return expr


@dispatch(object, object)
def compute_up(a, b, **kwargs):
    raise NotImplementedError("Blaze does not know how to compute "
                              "expression of type `%s` on data of type `%s`"
                              % (type(a).__name__, type(b).__name__))


@dispatch(Cast, object)
def compute_up(c, b, **kwargs):
    # cast only works on the expression system and does not affect the
    # computation
    return b


@dispatch(base)
def compute_up(a, **kwargs):
    return a


@dispatch((list, tuple))
def compute_up(seq, scope=None, **kwargs):
    return type(seq)(compute(item, scope or {}, **kwargs) for item in seq)


@dispatch(object)
def compute_down(expr, **kwargs):
    """ Compute the expression on the entire inputs

    inputs match up to leaves of the expression
    """
    return expr


def issubtype(a, b):
    """ A custom issubclass """
    if issubclass(a, b):
        return True
    if issubclass(a, (tuple, list, set)) and issubclass(b, Iterator):
        return True
    if issubclass(b, (tuple, list, set)) and issubclass(a, Iterator):
        return True
    return False


def type_change(old, new):
    """ Was there a significant type change between old and new data?

    >>> type_change([1, 2], [3, 4])
    False
    >>> type_change([1, 2], [3, [1,2,3]])
    True

    Some special cases exist, like no type change from list to Iterator

    >>> type_change([[1, 2]], [iter([1, 2])])
    False
    """
    if all(isinstance(x, base) for x in old + new):
        return False
    if len(old) != len(new):
        return True
    new_types = list(map(type, new))
    old_types = list(map(type, old))
    return not all(map(issubtype, new_types, old_types))


def top_then_bottom_then_top_again_etc(expr, scope, **kwargs):
    """ Compute expression against scope

    Does the following interpreter strategy:

    1.  Try compute_down on the entire expression
    2.  Otherwise compute_up from the leaves until we experience a type change
        (e.g. data changes from dict -> pandas DataFrame)
    3.  Re-optimize expression and re-pre-compute data
    4.  Go to step 1

    Examples
    --------

    >>> import numpy as np

    >>> s = symbol('s', 'var * {name: string, amount: int}')
    >>> data = np.array([('Alice', 100), ('Bob', 200), ('Charlie', 300)],
    ...                 dtype=[('name', 'S7'), ('amount', 'i4')])

    >>> e = s.amount.sum() + 1
    >>> top_then_bottom_then_top_again_etc(e, {s: data})
    601

    See Also
    --------

    bottom_up_until_type_break  -- uses this for bottom-up traversal
    top_to_bottom -- older version
    bottom_up -- older version still
    """
    # 0. Base case: expression is in dict, return associated data
    if expr in scope:
        return scope[expr]

    if not hasattr(expr, '_leaves'):
        return expr

    leaf_exprs = list(expr._leaves())
    leaf_data = [scope.get(leaf) for leaf in leaf_exprs]

    # 1. See if we have a direct computation path with compute_down
    try:
        return compute_down(expr, *leaf_data, **kwargs)
    except NotImplementedError:
        pass

    # 2. Compute from the bottom until there is a data type change
    expr2, scope2 = bottom_up_until_type_break(expr, scope, **kwargs)

    # 3. Re-optimize data and expressions
    optimize_ = kwargs.get('optimize', optimize)
    pre_compute_ = kwargs.get('pre_compute', pre_compute)
    if pre_compute_:
        scope3 = dict((e, pre_compute_(e, datum,
                                       **assoc(kwargs, 'scope', scope2)))
                      for e, datum in scope2.items())
    else:
        scope3 = scope2
    if optimize_:
        try:
            expr3 = optimize_(expr2, *[scope3[leaf]
                                       for leaf in expr2._leaves()])
            _d = dict(zip(expr2._leaves(), expr3._leaves()))
            scope4 = dict((e._subs(_d), d) for e, d in scope3.items())
        except NotImplementedError:
            expr3 = expr2
            scope4 = scope3
    else:
        expr3 = expr2
        scope4 = scope3

    # 4. Repeat
    if expr.isidentical(expr3):
        raise NotImplementedError("Don't know how to compute:\n"
                                  "type(expr): %s\n"
                                  "expr: %s\n"
                                  "data: %s" % (type(expr3), expr3, scope4))
    else:
        return top_then_bottom_then_top_again_etc(expr3, scope4, **kwargs)


_names = ('leaf_%d' % i for i in itertools.count(1))

_leaf_cache = {}
_used_tokens = defaultdict(set)


def _reset_leaves():
    _leaf_cache.clear()
    _used_tokens.clear()


def makeleaf(expr):
    """ Name of a new leaf replacement for this expression

    >>> _reset_leaves()

    >>> t = symbol('t', '{x: int, y: int, z: int}')
    >>> makeleaf(t) == t
    True
    >>> makeleaf(t.x)
    <`x` symbol; dshape='int32'>
    >>> makeleaf(t.x + 1)
    <`x` symbol; dshape='int64'>
    >>> makeleaf(t.y + 1)
    <`y` symbol; dshape='int64'>
    >>> makeleaf(t.x).isidentical(makeleaf(t.x + 1))
    False

    >>> from blaze import sin, cos
    >>> x = symbol('x', 'real')
    >>> makeleaf(cos(x)**2).isidentical(sin(x) ** 2)
    False

    >>> makeleaf(t) is t  # makeleaf passes on Symbols
    True
    """
    name = expr._name or '_'
    if expr in _leaf_cache:
        return _leaf_cache[expr]
    if isinstance(expr, Symbol):  # Idempotent on symbols
        _used_tokens[name].add(expr._token)
        _leaf_cache[expr] = expr
        return expr
    used_for_name = _used_tokens[name]
    for token in itertools.count():
        if token not in used_for_name:
            break
    result = symbol(name, expr.dshape, token)
    used_for_name.add(token)
    _leaf_cache[expr] = result
    return result


def data_leaves(expr, scope):
    return [scope[leaf] for leaf in expr._leaves()]


def bottom_up_until_type_break(expr, scope, **kwargs):
    """ Traverse bottom up until data changes significantly

    Parameters
    ----------

    expr: Expression
        Expression to compute
    scope: dict
        namespace matching leaves of expression to data

    Returns
    -------

    expr: Expression
        New expression with lower subtrees replaced with leaves
    scope: dict
        New scope with entries for those leaves

    Examples
    --------

    >>> import numpy as np

    >>> s = symbol('s', 'var * {name: string, amount: int}')
    >>> data = np.array([('Alice', 100), ('Bob', 200), ('Charlie', 300)],
    ...                 dtype=[('name', 'S7'), ('amount', 'i8')])

    This computation completes without changing type.  We get back a leaf
    symbol and a computational result

    >>> e = (s.amount + 1).distinct()
    >>> bottom_up_until_type_break(e, {s: data}) # doctest: +SKIP
    (amount, {amount: array([101, 201, 301])})

    This computation has a type change midstream (``list`` to ``int``), so we
    stop and get the unfinished computation.

    >>> e = s.amount.sum() + 1
    >>> bottom_up_until_type_break(e, {s: data})
    (amount_sum + 1, {<`amount_sum` symbol; dshape='int64'>: 600})
    """
    # 0. Base case.  Return if expression is in scope
    if expr in scope:
        leaf = makeleaf(expr)
        return leaf, {leaf: scope[expr]}

    inputs = list(unique(expr._inputs))

    # 1. Recurse down the tree, calling this function on children
    #    (this is the bottom part of bottom up)
    exprs, new_scopes = zip(*[bottom_up_until_type_break(i, scope, **kwargs)
                              for i in inputs])

    # 2. Form new (much shallower) expression and new (more computed) scope
    new_scope = toolz.merge(new_scopes)
    new_expr = expr._subs({
        i: e for i, e in zip(inputs, exprs) if not i.isidentical(e)
    })

    old_expr_leaves = expr._leaves()
    old_data_leaves = [scope.get(leaf) for leaf in old_expr_leaves]

    # 3. If the leaves have changed substantially then stop
    key = lambda x: str(type(x))
    if type_change(sorted(new_scope.values(), key=key),
                   sorted(old_data_leaves, key=key)):
        return new_expr, new_scope
    # 4. Otherwise try to do some actual work
    try:
        leaf = makeleaf(expr)
        _data = [new_scope[i] for i in new_expr._inputs]
    except KeyError:
        return new_expr, new_scope
    try:
        return leaf, {leaf: compute_up(new_expr, *_data, scope=new_scope,
                                       **kwargs)}
    except NotImplementedError:
        return new_expr, new_scope


def swap_resources_into_scope(expr, scope):
    """ Translate interactive expressions into normal abstract expressions

    Interactive Blaze expressions link to data on their leaves.  From the
    expr/compute perspective, this is a hack.  We push the resources onto the
    scope and return simple unadorned expressions instead.

    Examples
    --------

    >>> from blaze import data
    >>> t = data([1, 2, 3], dshape='3 * int', name='t')
    >>> swap_resources_into_scope(t.head(2), {})
    (t.head(2), {<`t` symbol; dshape='3 * int32'>: [1, 2, 3]})

    >>> expr, scope = _
    >>> list(scope.keys())[0]._resources()
    {}
    """
    resources = expr._resources()
    symbol_dict = dict((t, symbol(t._name, t.dshape)) for t in resources)
    resources = dict((symbol_dict[k], v) for k, v in resources.items())
    other_scope = dict((k, v) for k, v in scope.items()
                       if k not in symbol_dict)
    new_scope = toolz.merge(resources, other_scope)
    expr = expr._subs(symbol_dict)

    return expr, new_scope


@dispatch(Expr, Mapping)
def compute(expr, d, return_type=no_default, **kwargs):
    """Compute expression against data sources.

    Parameters
    ----------
    expr : Expr
        The blaze expression to compute.
    d : any
        The data source to compute expression on.
    return_type : {'native', 'core', type}, optional
        Type to return data as. Defaults to 'native' but will be changed
        to 'core' in version 0.11.  'core' forces the computation into a core
        type. 'native' returns the result as is from the respective backend's
        ``post_compute``. If a type is passed, it will odo the result into the
        type before returning.

    Examples
    --------
    >>> t = symbol('t', 'var * {name: string, balance: int}')
    >>> deadbeats = t[t['balance'] < 0]['name']

    >>> data = [['Alice', 100], ['Bob', -50], ['Charlie', -20]]
    >>> list(compute(deadbeats, {t: data}))
    ['Bob', 'Charlie']
    """
    _reset_leaves()
    optimize_ = kwargs.get('optimize', optimize)
    pre_compute_ = kwargs.get('pre_compute', pre_compute)
    post_compute_ = kwargs.get('post_compute', post_compute)
    expr2, d2 = swap_resources_into_scope(expr, d)
    if pre_compute_:
        d3 = dict(
            (e, pre_compute_(e, dat, **kwargs))
            for e, dat in d2.items()
            if e in expr2
        )
    else:
        d3 = d2

    if optimize_:
        try:
            expr3 = optimize_(expr2, *[v for e, v in d3.items() if e in expr2])
            _d = dict(zip(expr2._leaves(), expr3._leaves()))
            d4 = dict((e._subs(_d), d) for e, d in d3.items())
        except NotImplementedError:
            expr3 = expr2
            d4 = d3
    else:
        expr3 = expr2
        d4 = d3

    result = top_then_bottom_then_top_again_etc(expr3, d4, **kwargs)
    if post_compute_:
        result = post_compute_(expr3, result, scope=d4)

    # return the backend's native response
    if return_type is no_default:
        msg = ("The default behavior of compute will change in version >= 0.11"
               " where the `return_type` parameter will default to 'core'.")
        warnings.warn(msg, DeprecationWarning)
    # return result as a core type
    # (python type, pandas Series/DataFrame, numpy array)
    elif return_type == 'core':
        result = coerce_core(result, expr.dshape)
    # user specified type
    elif isinstance(return_type, type):
        result = into(return_type, result, dshape=expr3.dshape)
    elif return_type != 'native':
        raise ValueError(
            "Invalid return_type passed to compute: {}".format(return_type),
        )

    return result


@compute.register(Expr, object)
def compute_single_object(expr, o, **kwargs):
    """ Compute against single input

    Assumes that only one Symbol exists in expression

    >>> t = symbol('t', 'var * {name: string, balance: int}')
    >>> deadbeats = t[t['balance'] < 0]['name']

    >>> data = [['Alice', 100], ['Bob', -50], ['Charlie', -20]]
    >>> # list(compute(deadbeats, {t: data}))
    >>> list(compute(deadbeats, data))
    ['Bob', 'Charlie']
    """
    ts = set([x for x in expr._subterms() if isinstance(x, Symbol)])
    if len(ts) == 1:
        return compute(expr, {first(ts): o}, **kwargs)
    else:
        raise ValueError("Give compute dictionary input, got %s" % str(o))


@dispatch(Field, Mapping)
def compute_up(expr, data, **kwargs):
    return data[expr._name]


@compute_up.register(Join, object, object)
def join_dataframe_to_selectable(expr, lhs, rhs, scope=None, **kwargs):
    lexpr, rexpr = expr._leaves()
    return compute(
        expr,
        {
            lexpr: odo(lhs, pd.DataFrame, dshape=lexpr.dshape),
            rexpr: odo(rhs, pd.DataFrame, dshape=rexpr.dshape)
        },
        **kwargs
    )



from __future__ import absolute_import, division, print_function

from functools import partial
from multipledispatch import Dispatcher

__all__ = 'select_functions',

def match(condition, data):
    """ Does the condition match the data

    >>> match(1, 1)
    True
    >>> match(int, 1)
    True
    >>> match(lambda x: x > 0, 1)
    True

    Use tuples for many possibilities

    >>> match((1, 2, 3), 1)
    True
    >>> match((int, float), 1)
    True
    """
    return ((condition == data) or
            (isinstance(condition, type)
                and isinstance(data, condition)) or
            (not isinstance(condition, type)
                and callable(condition)
                and condition(data)) or
            (isinstance(condition, tuple)
                and any(match(c, data) for c in condition)))


def name(func):
    """ Name of a function, robust to partials """
    if isinstance(func, partial):
        return name(func.func)
    elif isinstance(func, Dispatcher):
        return func.name
    else:
        return func.__name__


def select_functions(methods, data):
    """
    Select appropriate functions given types and predicates
    """
    s = set()
    for condition, funcs in methods:
        if match(condition, data):
            s |= funcs
    return dict((name(func), func) for func in s)



"""

>>> from blaze import *

>>> accounts = symbol('accounts', 'var * {name: string, amount: int}')
>>> deadbeats = accounts[accounts['amount'] < 0]['name']

>>> from sqlalchemy import Table, Column, MetaData, Integer, String
>>> t = Table('accounts', MetaData(),
...           Column('name', String, primary_key = True),
...           Column('amount', Integer))
>>> print(compute(deadbeats, t))  # doctest: +SKIP
SELECT accounts.name
FROM accounts
WHERE accounts.amount < :amount_1
"""

from __future__ import absolute_import, division, print_function

from copy import copy
import datetime
import itertools
from itertools import chain
from operator import and_, eq, attrgetter
import warnings

from datashape import TimeDelta, Option, int32
from datashape.predicates import iscollection, isscalar, isrecord
import numpy as np
import numbers
from odo.backends.sql import metadata_of_engine, dshape_to_alchemy
from multipledispatch import MDNotImplementedError
import sqlalchemy as sa
from sqlalchemy import sql, Table, MetaData
from sqlalchemy.engine import Engine
from sqlalchemy.sql import Selectable, Select, functions as safuncs
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.elements import ClauseElement, ColumnElement, ColumnClause
from sqlalchemy.sql.selectable import FromClause, ScalarSelect
import toolz
from toolz import unique, concat, pipe, first
from toolz.compatibility import zip
from toolz.curried import map

from .core import compute_up, compute, base
from ..compatibility import reduce, basestring, _inttypes
from ..dispatch import dispatch
from ..expr import (
    BinOp,
    BinaryMath,
    Broadcast,
    By,
    Coalesce,
    Coerce,
    Concat,
    DateTime,
    DateTimeTruncate,
    Distinct,
    Expr,
    Field,
    FloorDiv,
    Head,
    IsIn,
    Join,
    Label,
    Like,
    Merge,
    Pow,
    Projection,
    ReLabel,
    Reduction,
    Sample,
    Selection,
    Shift,
    Slice,
    Sort,
    Sub,
    Summary,
    Tail,
    UnaryOp,
    UnaryStringFunction,
    common_subexpression,
    count,
    greatest,
    least,
    mean,
    nelements,
    notnull,
    nunique,
    reductions,
    std,
    StrCat,
    StrFind,
    StrSlice,
    var,
)
from ..expr.strings import len as str_len
from ..expr.broadcast import broadcast_collect
from ..expr.math import isnan
from ..utils import listpack


__all__ = ['sa', 'select']


def inner_columns(s):
    try:
        return s.inner_columns
    except AttributeError:
        return s.columns


@dispatch(Projection, Select)
def compute_up(expr, data, **kwargs):
    d = dict((c.name, c) for c in getattr(data, 'inner_columns', data.c))
    return data.with_only_columns([d[field] for field in expr.fields])


@dispatch(Projection, Selectable)
def compute_up(expr, data, **kwargs):
    return compute(
        expr,
        sa.select([data]),
        post_compute=False,
        return_type='native',
    )


@dispatch(Projection, sa.Column)
def compute_up(expr, data, **kwargs):
    selectables = [
        compute(
            expr._child[field],
            data,
            post_compute=False,
            return_type='native',
        )
        for field in expr.fields
    ]
    froms = set(concat(s.froms for s in selectables))
    assert 1 <= len(froms) <= 2, 'len(froms) = %s' % len(froms)
    result = unify_froms(sa.select(first(sel.inner_columns)
                                   for sel in selectables),
                         froms)
    return result.where(unify_wheres(selectables))


@dispatch(Projection, Select)
def compute_up(expr, data, **kwargs):
    return data.with_only_columns(
        first(compute(
            expr._child[field],
            data,
            post_compute=False,
            return_type='native',
        ).inner_columns)
        for field in expr.fields
    )


@dispatch(Field, FromClause)
def compute_up(t, s, **kwargs):
    return s.c[t._name]


def unify_froms(select, selectables):
    return reduce(lambda x, y: x.select_from(y), selectables, select)


def unify_wheres(selectables):
    clauses = list(unique((s._whereclause for s in selectables
                           if hasattr(s, '_whereclause')), key=str))
    return reduce(and_, clauses) if clauses else None


@dispatch(Field, Select)
def compute_up(expr, data, **kwargs):
    name = expr._name
    try:
        inner_columns = list(data.inner_columns)
        names = list(c.name for c in data.inner_columns)
        column = inner_columns[names.index(name)]
    except (KeyError, ValueError):
        single_column_select = compute(
            expr,
            first(data.inner_columns),
            post_compute=False,
            return_type='native',
        )
        column = first(single_column_select.inner_columns)
        result = unify_froms(sa.select([column]),
                             data.froms + single_column_select.froms)
        return result.where(unify_wheres([data, single_column_select]))
    else:
        return data.with_only_columns([column])


@dispatch(Field, sa.Column)
def compute_up(t, s, **kwargs):
    assert len(s.foreign_keys) == 1, 'exactly one foreign key allowed'
    key_col = first(s.foreign_keys).column
    return sa.select([key_col.table.c[t._name]]).where(s == key_col)


@dispatch(Broadcast, Select)
def compute_up(t, s, **kwargs):
    cols = list(inner_columns(s))
    d = dict((t._scalars[0][c], cols[i])
             for i, c in enumerate(t._scalars[0].fields))
    result = compute(
        t._scalar_expr,
        d,
        post_compute=False,
        return_type='native',
    ).label(t._name)

    s = copy(s)
    s.append_column(result)
    return s.with_only_columns([result])


@dispatch(Broadcast, Selectable)
def compute_up(t, s, **kwargs):
    cols = list(inner_columns(s))
    d = dict((t._scalars[0][c], cols[i])
             for i, c in enumerate(t._scalars[0].fields))
    return compute(
        t._scalar_expr,
        d,
        post_compute=False,
        return_type='native',
    ).label(t._name)


@dispatch(Concat, (Select, Selectable), (Select, Selectable))
def compute_up(t, lhs, rhs, **kwargs):
    if t.axis != 0:
        raise ValueError(
            'Cannot concat along a non-zero axis in sql; perhaps you want'
            " 'merge'?",
        )

    return select(lhs).union_all(select(rhs)).alias()


@dispatch(Broadcast, ColumnElement)
def compute_up(t, s, **kwargs):
    expr = t._scalar_expr
    return compute(
        expr,
        s,
        post_compute=False,
        return_type='native',
    ).label(expr._name)


def _binop(type_, f):
    @dispatch(type_, ColumnElement)
    def compute_up(t, data, **kwargs):
        if isinstance(t.lhs, Expr):
            return t.op(data, t.rhs)
        else:
            return f(t, t.lhs, data)

    @dispatch(type_, Select)
    def compute_up(t, data, **kwargs):
        assert len(data.c) == 1, (
            'Select cannot have more than a single column when doing'
            ' arithmetic'
        )
        column = first(data.inner_columns)
        if isinstance(t.lhs, Expr):
            return f(t, column, t.rhs)
        else:
            return f(t, t.lhs, column)

    @compute_up.register(type_,
                         (Select, ColumnElement, base),
                         (Select, ColumnElement))
    @compute_up.register(type_,
                         (Select, ColumnElement),
                         base)
    def binop_sql(t, lhs, rhs, **kwargs):
        if isinstance(lhs, Select):
            assert len(lhs.c) == 1, (
                'Select cannot have more than a single column when doing'
                ' arithmetic, got %r' % lhs
            )
            lhs = first(lhs.inner_columns)
        if isinstance(rhs, Select):
            assert len(rhs.c) == 1, (
                'Select cannot have more than a single column when doing'
                ' arithmetic, got %r' % rhs
            )
            rhs = first(rhs.inner_columns)

        return f(t, lhs, rhs)


_binop(BinOp, lambda expr, lhs, rhs: expr.op(lhs, rhs))
_binop(
    (greatest, least),
    lambda expr, lhs, rhs: getattr(sa.func, type(expr).__name__)(lhs, rhs),
)


@dispatch(Pow, ColumnElement)
def compute_up(t, data, **kwargs):
    if isinstance(t.lhs, Expr):
        return sa.func.pow(data, t.rhs)
    else:
        return sa.func.pow(t.lhs, data)


@dispatch(Pow, Select)
def compute_up(t, data, **kwargs):
    assert len(data.c) == 1, (
        'Select cannot have more than a single column when doing'
        ' arithmetic, got %r' % data
    )
    column = first(data.inner_columns)
    if isinstance(t.lhs, Expr):
        return sa.func.pow(column, t.rhs)
    else:
        return sa.func.pow(t.lhs, column)


@compute_up.register(Pow, (ColumnElement, base), ColumnElement)
@compute_up.register(Pow, ColumnElement, base)
def binop_sql_pow(t, lhs, rhs, **kwargs):
    return sa.func.pow(lhs, rhs)


@dispatch(BinaryMath, ColumnElement)
def compute_up(t, data, **kwargs):
    op = getattr(sa.func, type(t).__name__)
    if isinstance(t.lhs, Expr):
        return op(data, t.rhs)
    else:
        return op(t.lhs, data)


@dispatch(BinaryMath, Select)
def compute_up(t, data, **kwargs):
    assert len(data.c) == 1, (
        'Select cannot have more than a single column when doing'
        ' arithmetic, got %r' % data
    )
    column = first(data.inner_columns)
    op = getattr(sa.func, type(t).__name__)
    if isinstance(t.lhs, Expr):
        return op(column, t.rhs)
    else:
        return op(t.lhs, column)


@compute_up.register(BinaryMath, (ColumnElement, base), ColumnElement)
@compute_up.register(BinaryMath, ColumnElement, base)
def binary_math_sql(t, lhs, rhs, **kwargs):
    return getattr(sa.func, type(t).__name__)(lhs, rhs)


@compute_up.register(BinaryMath, Select, base)
def binary_math_sql_select(t, lhs, rhs, **kwargs):
    left, right = first(lhs.inner_columns), rhs
    result = getattr(sa.func, type(t).__name__)(left, right)
    return reconstruct_select([result], lhs)


@compute_up.register(BinaryMath, base, Select)
def binary_math_sql_select(t, lhs, rhs, **kwargs):
    left, right = lhs, first(rhs.inner_columns)
    result = getattr(sa.func, type(t).__name__)(left, right)
    return reconstruct_select([result], rhs)


@compute_up.register(BinaryMath, Select, Select)
def binary_math_sql_select(t, lhs, rhs, **kwargs):
    left, right = first(lhs.inner_columns), first(rhs.inner_columns)
    result = getattr(sa.func, type(t).__name__)(left, right)
    assert lhs.table == rhs.table, '%s != %s' % (lhs.table, rhs.table)
    return reconstruct_select([result], lhs.table)


@dispatch(FloorDiv, ColumnElement)
def compute_up(t, data, **kwargs):
    if isinstance(t.lhs, Expr):
        return sa.func.floor(data / t.rhs)
    else:
        return sa.func.floor(t.rhs / data)


@compute_up.register(FloorDiv, (ColumnElement, base), ColumnElement)
@compute_up.register(FloorDiv, ColumnElement, base)
def binop_sql(t, lhs, rhs, **kwargs):
    return sa.func.floor(lhs / rhs)


@dispatch(isnan, ColumnElement)
def compute_up(t, s, **kwargs):
    return s == float('nan')


@dispatch(UnaryOp, ColumnElement)
def compute_up(t, s, **kwargs):
    sym = t.symbol
    return getattr(t, 'op', getattr(safuncs, sym, getattr(sa.func, sym)))(s)


@dispatch(Selection, sa.sql.ColumnElement)
def compute_up(expr, data, scope=None, **kwargs):
    predicate = compute(
        expr.predicate,
        data,
        post_compute=False,
        return_type='native',
    )
    return compute(
        expr,
        {expr._child: data, expr.predicate: predicate},
        return_type='native',
        **kwargs
    )


@dispatch(Selection, sa.sql.ColumnElement, ColumnElement)
def compute_up(expr, col, predicate, **kwargs):
    return sa.select([col]).where(predicate)


@dispatch(Selection, Selectable)
def compute_up(expr, sel, scope=None, **kwargs):
    return compute(
        expr,
        {
            expr._child: sel,
            expr.predicate: compute(
                expr.predicate,
                toolz.merge(
                    {
                        expr._child[col.name]: col
                        for col in getattr(sel, 'inner_columns', sel.columns)
                    },
                    scope,
                ),
                optimize=False,
                post_compute=False,
            ),
        },
        return_type='native',
        **kwargs
    )


@dispatch(Selection, Selectable, ColumnElement)
def compute_up(expr, tbl, predicate, scope=None, **kwargs):
    try:
        return tbl.where(predicate)
    except AttributeError:
        return select([tbl]).where(predicate)


@dispatch(Selection, Selectable, Selectable)
def compute_up(expr, tbl, predicate, **kwargs):
    col, = inner_columns(predicate)
    return reconstruct_select(
        inner_columns(tbl),
        tbl,
        whereclause=unify_wheres((tbl, predicate)),
    ).where(col)


def select(s):
    """ Permissive SQL select

    Idempotent sa.select

    Wraps input in list if neccessary
    """
    if not isinstance(s, sa.sql.Select):
        if not isinstance(s, (tuple, list)):
            s = [s]
        s = sa.select(s)
    return s


table_names = ('table_%d' % i for i in itertools.count(1))


def name(sel):
    """ Name of a selectable """
    if hasattr(sel, 'name'):
        return sel.name
    if hasattr(sel, 'froms'):
        if len(sel.froms) == 1:
            return name(sel.froms[0])
    return next(table_names)


@dispatch(Select, Select)
def _join_selectables(a, b, condition=None, **kwargs):
    return a.join(b, condition, **kwargs)


@dispatch(Select, ClauseElement)
def _join_selectables(a, b, condition=None, **kwargs):
    if len(a.froms) > 1:
        raise MDNotImplementedError()

    return a.replace_selectable(a.froms[0],
                                a.froms[0].join(b, condition, **kwargs))


@dispatch(ClauseElement, Select)
def _join_selectables(a, b, condition=None, **kwargs):
    if len(b.froms) > 1:
        raise MDNotImplementedError()
    return b.replace_selectable(b.froms[0],
                                a.join(b.froms[0], condition, **kwargs))


@dispatch(ClauseElement, ClauseElement)
def _join_selectables(a, b, condition=None, **kwargs):
    return a.join(b, condition, **kwargs)


_getname = attrgetter('name')


def _clean_join_name(opposite_side_colnames, suffix, c):
    if c.name not in opposite_side_colnames:
        return c
    else:
        return c.label(c.name + suffix)


@dispatch(Join, ClauseElement, ClauseElement)
def compute_up(t, lhs, rhs, **kwargs):
    if isinstance(lhs, ColumnElement):
        lhs = select(lhs)
    if isinstance(rhs, ColumnElement):
        rhs = select(rhs)
    if name(lhs) == name(rhs):
        left_suffix, right_suffix = t.suffixes
        lhs = lhs.alias('%s%s' % (name(lhs), left_suffix))
        rhs = rhs.alias('%s%s' % (name(rhs), right_suffix))

    lhs = alias_it(lhs)
    rhs = alias_it(rhs)

    if isinstance(lhs, Select):
        lhs = lhs.alias(next(aliases))
        left_conds = [lhs.c.get(c) for c in listpack(t.on_left)]
    else:
        ldict = dict((c.name, c) for c in inner_columns(lhs))
        left_conds = [ldict.get(c) for c in listpack(t.on_left)]

    if isinstance(rhs, Select):
        rhs = rhs.alias(next(aliases))
        right_conds = [rhs.c.get(c) for c in listpack(t.on_right)]
    else:
        rdict = dict((c.name, c) for c in inner_columns(rhs))
        right_conds = [rdict.get(c) for c in listpack(t.on_right)]

    condition = reduce(and_, map(eq, left_conds, right_conds))

    # Perform join
    if t.how == 'inner':
        join = _join_selectables(lhs, rhs, condition=condition)
        main = lhs
    elif t.how == 'left':
        main, other = lhs, rhs
        join = _join_selectables(lhs, rhs, condition=condition, isouter=True)
    elif t.how == 'right':
        join = _join_selectables(rhs, lhs, condition=condition, isouter=True)
        main = rhs
    else:
        # http://stackoverflow.com/questions/20361017/sqlalchemy-full-outer-join
        raise ValueError("SQLAlchemy doesn't support full outer Join")

    """
    We now need to arrange the columns in the join to match the columns in
    the expression.  We care about order and don't want repeats
    """
    if isinstance(join, Select):
        def cols(x):
            if isinstance(x, Select):
                return list(x.inner_columns)
            else:
                return list(x.columns)
    else:
        cols = lambda x: list(x.columns)

    main_cols = cols(main)
    left_cols = cols(lhs)
    left_names = set(map(_getname, left_cols))
    right_cols = cols(rhs)
    right_names = set(map(_getname, right_cols))

    left_suffix, right_suffix = t.suffixes
    fields = [
        f.replace(left_suffix, '').replace(right_suffix, '') for f in t.fields
    ]
    columns = [c for c in main_cols if c.name in t._on_left]
    columns += [_clean_join_name(right_names, left_suffix, c)
                for c in left_cols
                if c.name in fields and c.name not in t._on_left]
    columns += [_clean_join_name(left_names, right_suffix, c)
                for c in right_cols
                if c.name in fields and c.name not in t._on_right]

    if isinstance(join, Select):
        return join.with_only_columns(columns)
    else:
        return sa.select(columns, from_obj=join)


names = {
    mean: 'avg'
}


def reconstruct_select(columns, original, **kwargs):
    return sa.select(columns,
                     from_obj=kwargs.pop('from_obj',
                                         getattr(original, '_from_obj', None)),
                     whereclause=kwargs.pop('whereclause',
                                            getattr(original,
                                                    '_whereclause', None)),
                     bind=kwargs.pop('bind', original.bind),
                     distinct=kwargs.pop('distinct',
                                         getattr(original,
                                                 '_distinct', False)),
                     group_by=kwargs.pop('group_by',
                                         getattr(original,
                                                 '_group_by_clause', None)),
                     having=kwargs.pop('having',
                                       getattr(original, '_having', None)),
                     limit=kwargs.pop('limit',
                                      getattr(original, '_limit', None)),
                     offset=kwargs.pop('offset',
                                       getattr(original, '_offset', None)),
                     order_by=kwargs.pop('order_by',
                                         getattr(original,
                                                 '_order_by_clause', None)),
                     **kwargs)


@dispatch((nunique, Reduction), Select)
def compute_up(expr, data, **kwargs):
    if expr.axis != (0,):
        raise ValueError('axis not equal to 0 not defined for SQL reductions')
    data = data.alias(name=next(aliases))
    cols = list(inner_columns(data))
    d = dict((expr._child[c], cols[i])
             for i, c in enumerate(expr._child.fields))
    return select([
        compute(expr, d, post_compute=False, return_type='native')
    ])


@dispatch(Distinct, ColumnElement)
def compute_up(t, s, **kwargs):
    return s.distinct(*t.on).label(t._name)


@dispatch(Distinct, Select)
def compute_up(t, s, **kwargs):
    return s.distinct(*t.on)


@dispatch(Distinct, Selectable)
def compute_up(t, s, **kwargs):
    return select(s).distinct(*t.on)


@dispatch(Reduction, ClauseElement)
def compute_up(t, s, **kwargs):
    if t.axis != (0,):
        raise ValueError('axis not equal to 0 not defined for SQL reductions')
    try:
        op = getattr(sa.sql.functions, t.symbol)
    except AttributeError:
        op = getattr(sa.sql.func, names.get(type(t), t.symbol))
    return op(s).label(t._name)


prefixes = {
    std: 'stddev',
    var: 'var'
}


@dispatch((std, var), sql.elements.ColumnElement)
def compute_up(t, s, **kwargs):
    measure = t.schema.measure
    is_timedelta = isinstance(getattr(measure, 'ty', measure), TimeDelta)
    if is_timedelta:
        # part 1 of 2 to work around the fact that postgres does not have
        # timedelta var or std: cast to a double which is seconds
        s = sa.extract('epoch', s)
    if t.axis != (0,):
        raise ValueError('axis not equal to 0 not defined for SQL reductions')
    funcname = 'samp' if t.unbiased else 'pop'
    full_funcname = '%s_%s' % (prefixes[type(t)], funcname)
    ret = getattr(sa.func, full_funcname)(s)
    if is_timedelta:
        # part 2 of 2 to work around the fact that postgres does not have
        # timedelta var or std: cast back from seconds by
        # multiplying by a 1 second timedelta
        ret = ret * datetime.timedelta(seconds=1)
    return ret.label(t._name)


@dispatch(count, Selectable)
def compute_up(t, s, **kwargs):
    return s.count()


@dispatch(count, sa.Table)
def compute_up(t, s, **kwargs):
    if t.axis != (0,):
        raise ValueError('axis not equal to 0 not defined for SQL reductions')
    try:
        c = list(s.primary_key)[0]
    except IndexError:
        c = list(s.columns)[0]

    return sa.func.count(c)


@dispatch(nelements, (Select, ClauseElement))
def compute_up(t, s, **kwargs):
    return compute_up(t._child.count(), s)


@dispatch(count, Select)
def compute_up(t, s, **kwargs):
    if t.axis != (0,):
        raise ValueError('axis not equal to 0 not defined for SQL reductions')
    al = next(aliases)
    try:
        s2 = s.alias(al)
        col = list(s2.primary_key)[0]
    except (KeyError, IndexError):
        s2 = s.alias(al)
        col = list(s2.columns)[0]

    result = sa.func.count(col)

    return select([list(inner_columns(result))[0].label(t._name)])


@dispatch(nunique, (sa.sql.elements.Label, sa.Column))
def compute_up(t, s, **kwargs):
    if t.axis != (0,):
        raise ValueError('axis not equal to 0 not defined for SQL reductions')
    return sa.func.count(s.distinct())


@dispatch(nunique, Selectable)
def compute_up(expr, data, **kwargs):
    return select(data).distinct().alias(next(aliases)).count()


@dispatch(By, sa.Column)
def compute_up(expr, data, scope=None, **kwargs):
    data = lower_column(data)
    grouper = compute(
        expr.grouper,
        scope,
        post_compute=False,
        return_type='native',
        **kwargs
    )

    app = expr.apply
    reductions = [
        compute(
            val,
            data,
            post_compute=None,
            return_type='native',
        ).label(name)
        for val, name in zip(app.values, app.fields)
    ]

    froms = list(unique(chain(get_all_froms(grouper),
                              concat(map(get_all_froms, reductions)))))
    inner_cols = list(getattr(grouper, 'inner_columns', [grouper]))
    grouper_cols = inner_cols[:]
    inner_cols.extend(concat(
        getattr(getattr(r, 'element', None), 'inner_columns', [r])
        for r in reductions
    ))
    wheres = unify_wheres([grouper] + reductions)
    sel = unify_froms(sa.select(inner_cols, whereclause=wheres), froms)
    return sel.group_by(*grouper_cols)


@dispatch(By, ClauseElement)
def compute_up(expr, data, **kwargs):
    if not valid_grouper(expr.grouper):
        raise TypeError("Grouper must have a non-nested record or one "
                        "dimensional collection datashape, "
                        "got %s of type %r with dshape %s" %
                        (expr.grouper, type(expr.grouper).__name__,
                         expr.grouper.dshape))
    grouper = get_inner_columns(
        compute(
            expr.grouper,
            data,
            post_compute=False,
            return_type='native',
        ),
    )
    app = expr.apply
    reductions = [
        compute(
            val,
            data,
            post_compute=False,
            return_type='native',
        ).label(name)
        for val, name in zip(app.values, app.fields)
    ]

    return sa.select(grouper + reductions).group_by(*grouper)


def lower_column(col):
    """ Return column from lower level tables if possible

    >>> metadata = sa.MetaData()

    >>> s = sa.Table('accounts', metadata,
    ...              sa.Column('name', sa.String),
    ...              sa.Column('amount', sa.Integer),
    ...              sa.Column('id', sa.Integer, primary_key=True),
    ...              )

    >>> s2 = select([s])
    >>> s2.c.amount is s.c.amount
    False

    >>> lower_column(s2.c.amount) is s.c.amount
    True

    >>> lower_column(s2.c.amount)
    Column('amount', Integer(), table=<accounts>)
    """

    old = None
    while col is not None and col is not old:
        old = col
        if not hasattr(col, 'table') or not hasattr(col.table, 'froms'):
            return col
        for f in col.table.froms:
            if f.corresponding_column(col) is not None:
                col = f.corresponding_column(col)

    return old


aliases = ('alias_%d' % i for i in itertools.count(1))


@toolz.memoize
def alias_it(s):
    """ Alias a Selectable if it has a group by clause """
    if (hasattr(s, '_group_by_clause') and
            s._group_by_clause is not None and
            len(s._group_by_clause)):
        return s.alias(next(aliases))
    else:
        return s


def is_nested_record(measure):
    """Predicate for checking whether `measure` is a nested ``Record`` dshape

    Examples
    --------
    >>> from datashape import dshape
    >>> is_nested_record(dshape('{a: int32, b: int32}').measure)
    False
    >>> is_nested_record(dshape('{a: var * ?float64, b: ?string}').measure)
    True
    """
    if not isrecord(measure):
        raise TypeError('Input must be a Record type got %s of type %r' %
                        (measure, type(measure).__name__))
    return not all(isscalar(getattr(t, 'key', t)) for t in measure.types)


def valid_grouper(expr):
    ds = expr.dshape
    measure = ds.measure
    return (iscollection(ds) and
            (isscalar(getattr(measure, 'key', measure)) or
             (isrecord(measure) and not is_nested_record(measure))))


def valid_reducer(expr):
    ds = expr.dshape
    measure = ds.measure
    return (not iscollection(ds) and
            (isscalar(measure) or
             (isrecord(measure) and not is_nested_record(measure))))


@dispatch(By, Select)
def compute_up(expr, data, **kwargs):
    if not valid_grouper(expr.grouper):
        raise TypeError("Grouper must have a non-nested record or one "
                        "dimensional collection datashape, "
                        "got %s of type %r with dshape %s" %
                        (expr.grouper, type(expr.grouper).__name__,
                         expr.grouper.dshape))

    s = alias_it(data)

    if valid_reducer(expr.apply):
        reduction = compute(
            expr.apply,
            s,
            post_compute=False, return_type='native',
        )
    else:
        raise TypeError('apply must be a Summary expression')

    grouper = get_inner_columns(compute(
        expr.grouper,
        s,
        post_compute=False,
        return_type='native',
    ))
    reduction_columns = pipe(reduction.inner_columns,
                             map(get_inner_columns),
                             concat)
    columns = list(unique(chain(grouper, reduction_columns)))
    if (not isinstance(s, sa.sql.selectable.Alias) or
            (hasattr(s, 'froms') and isinstance(s.froms[0],
                                                sa.sql.selectable.Join))):
        assert len(s.froms) == 1, 'only a single FROM clause supported for now'
        from_obj, = s.froms
    else:
        from_obj = None

    return reconstruct_select(columns,
                              getattr(s, 'element', s),
                              from_obj=from_obj,
                              group_by=grouper)


@dispatch(Sort, (Selectable, Select))
def compute_up(t, s, **kwargs):
    s = select(s.alias())
    direction = sa.asc if t.ascending else sa.desc
    cols = [direction(lower_column(s.c[c])) for c in listpack(t.key)]
    return s.order_by(*cols)


@dispatch(Sort, (sa.Table, ColumnElement))
def compute_up(t, s, **kwargs):
    s = select(s)
    direction = sa.asc if t.ascending else sa.desc
    cols = [direction(lower_column(s.c[c])) for c in listpack(t.key)]
    return s.order_by(*cols)


def _samp_compute_up(t, s, **kwargs):
    if t.n is not None:
        limit = t.n
    else:
        limit = sa.select([safuncs.count() * t.frac],
                          from_obj=s.alias()).as_scalar()
    return s.order_by(safuncs.random()).limit(limit)


@dispatch(Sample, sa.Table)
def compute_up(t, s, **kwargs):
    return _samp_compute_up(t, select(s), **kwargs)


@dispatch(Sample, ColumnElement)
def compute_up(t, s, **kwargs):
    return _samp_compute_up(t, sa.select([s]), **kwargs)


@dispatch(Sample, FromClause)
def compute_up(t, s, **kwargs):
    return _samp_compute_up(t, s, **kwargs)


@dispatch(Head, FromClause)
def compute_up(t, s, **kwargs):
    if s._limit is not None and s._limit <= t.n:
        return s
    return s.limit(t.n)


@dispatch(Head, sa.Table)
def compute_up(t, s, **kwargs):
    return s.select().limit(t.n)


@dispatch(Head, ColumnElement)
def compute_up(t, s, **kwargs):
    return sa.select([s]).limit(t.n)


@dispatch(Head, ScalarSelect)
def compute_up(t, s, **kwargs):
    return compute(t, s.element, post_compute=False, return_type='native')


@dispatch(Label, ColumnElement)
def compute_up(t, s, **kwargs):
    return s.label(t.label)


@dispatch(Label, FromClause)
def compute_up(t, s, **kwargs):
    assert len(s.c) == 1, \
        'expected %s to have a single column but has %d' % (s, len(s.c))
    inner_column, = s.inner_columns
    return reconstruct_select([inner_column.label(t.label)], s)


@dispatch(Expr, ScalarSelect)
def post_compute(t, s, **kwargs):
    return s.element


@dispatch(ReLabel, Selectable)
def compute_up(expr, data, **kwargs):
    names = data.c.keys()
    assert names == expr._child.fields, (
        'names = %r\nexpr._child.fields = %r' % (names, expr._child.fields)
    )
    d = dict(zip(names, getattr(data, 'inner_columns', data.c)))
    return reconstruct_select(
        (d[col].label(new_col) if col != new_col else d[col]
         for col, new_col in zip(expr._child.fields, expr.fields)),
        data,
    )


@dispatch(FromClause)
def get_inner_columns(sel):
    try:
        return list(sel.inner_columns)
    except AttributeError:
        return list(map(lower_column, sel.c.values()))


@dispatch(ColumnElement)
def get_inner_columns(c):
    return [c]


@dispatch(ScalarSelect)
def get_inner_columns(sel):
    inner_columns = list(sel.inner_columns)
    assert len(inner_columns) == 1, 'ScalarSelect should have only ONE column'
    return list(map(lower_column, inner_columns))


@dispatch(sa.sql.functions.Function)
def get_inner_columns(f):
    unique_columns = unique(concat(map(get_inner_columns, f.clauses)))
    lowered = [x.label(getattr(x, 'name', None)) for x in unique_columns]
    return [getattr(sa.func, f.name)(*lowered)]


@dispatch(sa.sql.elements.Label)
def get_inner_columns(label):
    """
    Notes
    -----
    This should only ever return a list of length 1

    This is because we need to turn ScalarSelects into an actual column
    """
    name = label.name
    inner_columns = get_inner_columns(label.element)
    assert len(inner_columns) == 1
    return [lower_column(c).label(name) for c in inner_columns]


@dispatch(Select)
def get_all_froms(sel):
    return list(unique(sel.locate_all_froms()))


@dispatch(sa.Table)
def get_all_froms(t):
    return [t]


@dispatch(sa.sql.elements.ColumnElement)
def get_all_froms(colelement):
    return list(unique(concat(map(get_all_froms, colelement.get_children()))))


@dispatch((ScalarSelect, sa.sql.elements.Label))
def get_all_froms(element):
    return get_all_froms(element.element)


@dispatch(sa.sql.functions.FunctionElement)
def get_all_froms(function):
    return list(unique(concat(map(get_all_froms, function.clauses.clauses))))


@dispatch(ColumnClause)
def get_all_froms(c):
    return [c.table]


def get_clause(data, kind):
    # arg SQLAlchemy doesn't allow things like data._group_by_clause or None
    assert kind == 'order_by' or kind == 'group_by', \
        'kind must be "order_by" or "group_by"'
    clause = getattr(data, '_%s_clause' % kind, None)
    return clause.clauses if clause is not None else None


@dispatch(Merge, (Selectable, Select, sa.Column))
def compute_up(expr, data, **kwargs):
    # get the common subexpression of all the children in the merge
    subexpression = common_subexpression(*expr.children)

    # compute each child, including the common subexpression
    children = [
        compute(
            child,
            {subexpression: data},
            post_compute=False,
            return_type='native',
        )
        for child in expr.children
    ]

    # Get the original columns from the selection and rip out columns from
    # Selectables and ScalarSelects
    columns = list(unique(concat(map(get_inner_columns, children))))

    # we need these getattrs if data is a ColumnClause or Table
    from_obj = get_all_froms(data)
    return reconstruct_select(columns, data, from_obj=from_obj)


@dispatch(Summary, Select)
def compute_up(t, s, scope=None, **kwargs):
    d = dict((t._child[c], list(inner_columns(s))[i])
             for i, c in enumerate(t._child.fields))

    cols = [
        compute(
            val,
            toolz.merge(scope, d),
            post_compute=None,
            return_type='native',
        ).label(name)
        for name, val in zip(t.fields, t.values)
    ]

    s = copy(s)
    for c in cols:
        s.append_column(c)

    return s.with_only_columns(cols)


@dispatch(Summary, ClauseElement)
def compute_up(t, s, **kwargs):
    scope = {t._child: s}
    return sa.select(
        compute(
            value,
            scope,
            post_compute=None,
            return_type='native',
        ).label(name)
        for value, name in zip(t.values, t.fields)
    )


@dispatch(Like, Select)
def compute_up(t, s, **kwargs):
    assert len(s.c) == 1, \
            'Select cannot have more than a single column when filtering with `like`'
    return compute_up(t, first(s.inner_columns), **kwargs)


@dispatch(Like, ColumnElement)
def compute_up(t, s, **kwargs):
    return s.like(t.pattern.replace('*', '%').replace('?', '_'))


# TODO: remove if the alternative fix goes into PyHive
@compiles(sa.sql.functions.Function, 'hive')
def compile_char_length_on_hive(element, compiler, **kwargs):
    assert len(element.clauses) == 1, \
        'char_length must have a single clause, got %s' % list(element.clauses)
    if element.name == 'char_length':
        return compiler.visit_function(sa.func.length(*element.clauses),
                                       **kwargs)
    return compiler.visit_function(element, **kwargs)


@dispatch(str_len, ColumnElement)
def compute_up(expr, data, **kwargs):
    return sa.sql.functions.char_length(data).label(expr._name)


@dispatch(StrSlice, ColumnElement)
def compute_up(expr, data, **kwargs):
    if isinstance(expr.slice, _inttypes):
        idx = expr.slice + 1
        if idx < 1: # SQL string indexing is 1-based and positive.
            msg = "Index {} out-of-bounds for SQL string indexing."
            raise IndexError(msg.format(expr.slice))
        args = idx, 1
    elif isinstance(expr.slice, tuple):
        start, stop, step = expr.slice
        if step is not None:
            msg = "step value {} not valid for SQL string indexing."
            raise ValueError(msg.format(step))
        norm_start = start if isinstance(start, _inttypes) else 0
        if norm_start < 0:
            msg = "Negative indexing not valid for SQL strings; given {}."
            raise ValueError(msg.format(norm_start))
        if isinstance(stop, _inttypes):
            if stop < 0:
                msg = "Negative indexing not valid for SQL strings; given {}."
                raise ValueError(msg.format(stop))
            args = norm_start + 1, (stop - norm_start)
        elif stop is None:
            args = norm_start + 1,
    return sa.sql.func.substring(data, *args)


@dispatch(StrFind, ColumnElement)
def compute_up(expr, data, **kwargs):
    sub = sa.sql.expression.literal(expr.sub)
    return sa.sql.func.position(sub.op('in')(data))


@compute_up.register(StrCat, Select, basestring)
@compute_up.register(StrCat, basestring, Select)
def str_cat_sql(expr, lhs, rhs, **kwargs):
    if isinstance(lhs, Select):
        orig = lhs
        lhs = first(lhs.inner_columns)
    else:
        orig = rhs
        rhs = first(rhs.inner_columns)
    if expr.sep:
        result = (lhs + expr.sep + rhs).label(expr.lhs._name)
    else:
        result = (lhs + rhs).label(expr.lhs._name)
    return reconstruct_select([result], orig)


@compute_up.register(StrCat, Select, Select)
def str_cat_sql(expr, lhs, rhs, **kwargs):
    left, right = first(lhs.inner_columns), first(rhs.inner_columns)
    if expr.sep:
        result = (left + expr.sep + right).label(expr.lhs._name)
    else:
        result = (left + right).label(expr.lhs._name)
    return reconstruct_select([result], lhs)


@compute_up.register(StrCat, (ColumnElement, basestring), ColumnElement)
@compute_up.register(StrCat, ColumnElement, basestring)
def str_cat_sql(expr, lhs, rhs, **kwargs):
    if expr.sep:
        return (lhs + expr.sep + rhs).label(expr.lhs._name)
    else:
        return (lhs + rhs).label(expr.lhs._name)


@dispatch(UnaryStringFunction, ColumnElement)
def compute_up(expr, data, **kwargs):
    func_name = type(expr).__name__
    return getattr(sa.sql.func, func_name)(data).label(expr._name)


@dispatch(notnull, ColumnElement)
def compute_up(expr, data, **kwargs):
    return data != None


@toolz.memoize
def table_of_metadata(metadata, name):
    if metadata.schema is not None:
        name = '.'.join((metadata.schema, name))
    if name not in metadata.tables:
        metadata.reflect(views=metadata.bind.dialect.supports_views)
    return metadata.tables[name]


def table_of_engine(engine, name):
    metadata = metadata_of_engine(engine)
    return table_of_metadata(metadata, name)


@dispatch(Field, sa.engine.Engine)
def compute_up(expr, data, **kwargs):
    return table_of_engine(data, expr._name)


@dispatch(DateTime, ColumnElement)
def compute_up(expr, data, **kwargs):
    if expr.attr == 'date':
        return sa.func.date(data).label(expr._name)

    return sa.extract(expr.attr, data).label(expr._name)


@dispatch(DateTimeTruncate, ColumnElement)
def compute_up(expr, data, **kwargs):
    return sa.func.date_trunc(expr.unit, data).label(expr._name)


@compiles(sa.sql.elements.Extract, 'hive')
def hive_extract_to_date_function(element, compiler, **kwargs):
    func = getattr(sa.func, element.field)(element.expr)
    return compiler.visit_function(func, **kwargs)


@compiles(sa.sql.elements.Extract, 'mssql')
def mssql_extract_to_datepart(element, compiler, **kwargs):
    func = sa.func.datepart(sa.sql.expression.column(element.field),
                            element.expr)
    return compiler.visit_function(func, **kwargs)


def engine_of(x):
    if isinstance(x, Engine):
        return x
    if isinstance(x, MetaData):
        return x.bind
    if isinstance(x, Table):
        return x.metadata.bind
    raise NotImplementedError("Can't deterimine engine of %s" % x)


@dispatch(object)
def _subexpr_optimize(expr):
    return expr


@dispatch(Expr)
def _subexpr_optimize(expr):
    return type(expr)(*map(_subexpr_optimize, expr._args))


timedelta_ns = TimeDelta(unit='ns')


@dispatch(reductions.any)
def _subexpr_optimize(expr):
    if expr.axis != tuple(range(expr._child.ndim)):
        raise ValueError("Cannot perform 'any' over an axis: %s")
    if expr.keepdims:
        raise ValueError("Cannot perform 'any' with keepdims=True")

    return expr._child.coerce(Option(int32)
                              if isinstance(expr._child.dshape, Option) else
                              int32).sum() != 0


@dispatch(reductions.all)
def _subexpr_optimize(expr):
    if expr.axis != tuple(range(expr._child.ndim)):
        raise ValueError("Cannot perform 'all' over an axis: %s")
    if expr.keepdims:
        raise ValueError("Cannot perform 'all' with keepdims=True")

    return (~expr._child).coerce(Option(int32)
                                 if isinstance(expr._child.dshape, Option) else
                                 int32).sum() == 0


@dispatch(Sub)
def _subexpr_optimize(expr):
    new_expr = type(expr)(*map(_subexpr_optimize, expr._args))
    schema = expr.schema
    # we have a timedelta shaped expression; sql timedeltas are in `ns` units
    # so we should coerce this exprssion over
    if isinstance(schema, TimeDelta) and schema.unit != 'ns':
        new_expr = new_expr.coerce(timedelta_ns)
    return new_expr


@dispatch(Tail)
def _subexpr_optimize(expr):
    child = sorter = expr._child
    while not isinstance(sorter, Sort):
        try:
            sorter = sorter._child
        except AttributeError:
            break
    else:
        # Invert the sort order, then take the head, then re-sort based on
        # the original key.
        return child._subs({
            sorter: sorter._child.sort(
                sorter._key,
                ascending=not sorter.ascending,
            ),
        }).head(expr.n).sort(sorter._key, ascending=sorter.ascending)

    # If there is no sort order, then we can swap out a head with a tail.
    # This is equivalent in this backend and considerably faster.
    warnings.warn(
        "'tail' of a sql operation with no sort is the same as 'head'",
    )
    return child.head(expr.n)


@dispatch(Expr, ClauseElement)
def optimize(expr, _):
    collected = broadcast_collect(expr, no_recurse=Selection)
    return reduce(
        lambda expr, term: expr._subs({term: _subexpr_optimize(term)}),
        collected._subterms(),
        collected,
    )


@dispatch(Field, sa.MetaData)
def compute_up(expr, data, **kwargs):
    return table_of_metadata(data, expr._name)


@dispatch(Expr, ClauseElement)
def post_compute(_, s, **kwargs):
    return select(s)


@dispatch(IsIn, ColumnElement)
def compute_up(expr, data, **kwargs):
    return data.in_(expr._keys)


@dispatch(IsIn, Selectable)
def compute_up(expr, data, **kwargs):
    assert len(data.columns) == 1, (
        'only 1 column is allowed in a Select in IsIn'
    )
    col, = inner_columns(data)
    return reconstruct_select((col.in_(expr._keys),), data)


@dispatch(Slice, (Select, Selectable, ColumnElement))
def compute_up(expr, data, **kwargs):
    index = expr.index[0]  # [0] replace_slices returns tuple ((start, stop), )
    if isinstance(index, slice):
        start = index.start or 0
        if start < 0:
            raise ValueError('start value of slice cannot be negative'
                             ' with a SQL backend')

        stop = index.stop
        if stop is not None and stop < 0:
            raise ValueError('stop value of slice cannot be negative with a '
                             'SQL backend.')

        if index.step is not None and index.step != 1:
            raise ValueError('step parameter in slice objects not supported '
                             'with SQL backend')

    elif isinstance(index, (np.integer, numbers.Integral)):
        if index < 0:
            raise ValueError('integer slice cannot be negative for the'
                             ' SQL backend')
        start = index
        stop = start + 1
    else:
        raise TypeError('type %r not supported for slicing wih SQL backend'
                        % type(index).__name__)

    warnings.warn('The order of the result set from a Slice expression '
                  'computed against the SQL backend is not deterministic.')

    if stop is None:  # Represents open-ended slice. e.g. [3:]
        return select(data).offset(start)
    else:
        return select(data).offset(start).limit(stop - start)


@dispatch(Coerce, ColumnElement)
def compute_up(expr, data, **kwargs):
    return sa.cast(data, dshape_to_alchemy(expr.to)).label(expr._name)


@dispatch(Coerce, Select)
def compute_up(expr, data, **kwargs):
    column = first(data.inner_columns)
    cast = sa.cast(column, dshape_to_alchemy(expr.to)).label(expr._name)
    return reconstruct_select([cast], data)


@dispatch(Shift, ColumnElement)
def compute_up(expr, data, **kwargs):
    return sa.func.lag(data, expr.n).over().label(expr._name)


@compute_up.register(Coalesce, (ColumnElement, base), ColumnElement)
@compute_up.register(Coalesce, ColumnElement, base)
def coalesce_sql(t, lhs, rhs, **kwargs):
    return sa.sql.functions.coalesce(lhs, rhs).label(t._name)


@compute_up.register(Coalesce, Select, base)
@compute_up.register(Coalesce, base, Select)
def coalesce_sql_select(expr, lhs, rhs, **kwargs):
    if isinstance(lhs, Select):
        orig = lhs
        lhs = first(lhs.inner_columns)
    else:
        orig = rhs
        rhs = first(rhs.inner_columns)
    result = sa.sql.functions.coalesce(lhs, rhs).label(expr._name)
    return reconstruct_select([result], orig)


@dispatch(Coalesce, (Select, ColumnElement))
def compute_up(expr, data, **kwargs):
    if isinstance(expr.lhs, Expr):
        lhs = data
        rhs = expr.rhs
    else:
        lhs = expr.lhs
        rhs = expr.data

    return compute_up(expr, lhs, rhs, **kwargs)



from __future__ import absolute_import, division, print_function

import pytest
import sys

from blaze.expr.arithmetic import (scalar_coerce, Mult, Add, dshape)
from blaze.expr.math import sin, cos, isnan, exp, log
from blaze.expr import symbol, eval_str
from blaze.compatibility import xfail, basestring
from datetime import date, datetime

x = symbol('x', 'real')
y = symbol('y', 'real')
b = symbol('b', 'bool')


def test_basic():
    expr = (x + y) * 3

    assert eval(str(expr)).isidentical(expr)
    assert expr.isidentical(
        Mult(Add(symbol('x', 'real'), symbol('y', 'real')), 3),
    )


def test_eval_str():
    expr = (x + y) * 3
    assert eval_str(expr) == '(x + y) * 3'

    assert eval_str(1) == '1'
    assert eval_str('Alice') == "'Alice'"
    assert "'Alice'" in eval_str(u'Alice')

    print(eval_str(-x))
    assert eval_str(-x) == '-x'

    assert '~' in eval_str(~b)


def test_str():
    x = symbol('x', 'real')
    assert str(x + 10) == 'x + 10'


def test_invert():
    x = symbol('x', 'bool')
    expr = ~x
    assert expr.op(x).isidentical(expr)


def test_boolean_math_has_boolean_methods():
    x = symbol('x', '?int')
    expr = ~(isnan(x)) | (x > 0)

    assert eval(str(expr)).isidentical(expr)


def ishashable(x):
    try:
        hash(x)
        return True
    except:
        return False


def test_Symbol_is_hashable():
    assert ishashable(x)


def test_relationals():
    x = symbol('x', 'real')
    for expr in [x < 1, x > 1, x == 1, x != 1, x <= 1, x >= 1, ~b]:
        assert expr.dshape == dshape('bool')
        assert eval(str(expr)).isidentical(expr)


def test_numbers():
    x = symbol('x', 'real')
    y = symbol('y', 'int')
    for expr in [x + 1, x - 1, x * 1, x + y, x - y, x / y, x * y + x + y,
                 x**y, x**2, 2**x, x % 5, -x,
                 sin(x), cos(x ** 2), exp(log(y))]:
        assert expr.dshape == dshape('real')
        assert eval(str(expr)).isidentical(expr)

    assert (-y).dshape == dshape('int')


@xfail(reason="TODO")
def test_neg_dshape_unsigned():
    y = symbol('y', 'uint32')
    assert (-y).dshape == dshape('int32')


def test_arithmetic_dshape_inference():
    x = symbol('x', 'int')
    y = symbol('y', 'int')
    assert (x + y).dshape == dshape('int')


def test_date_coercion():
    d = symbol('d', 'date')
    expr = d < '2012-01-01'
    assert isinstance(expr.rhs, date)


def test_datetime_coercion():
    d = symbol('d', 'datetime')
    expr = d > '2012-01-01T12:30:00'
    assert isinstance(expr.rhs, datetime)


def test_scalar_coerce():
    assert scalar_coerce('int', 1) == 1
    assert scalar_coerce('int', '1') == 1
    assert scalar_coerce('{x: int}', '1') == 1
    with pytest.raises(TypeError):
        scalar_coerce('{x: int, y: int}', '1')

    assert scalar_coerce('date', 'Jan 1st, 2012') == date(2012, 1, 1)

    assert (scalar_coerce('datetime', 'Jan 1st, 2012 12:00:00') ==
            datetime(2012, 1, 1, 12, 0, 0))

    with pytest.raises(TypeError):
        scalar_coerce('date', 'Jan 1st, 2012 12:00:00')

    assert scalar_coerce('?date', 'Jan 1st, 2012') == date(2012, 1, 1)
    assert scalar_coerce('?date', '2012-12-01') == date(2012, 12, 1)
    with pytest.raises(TypeError):
        scalar_coerce('?date', '')
    assert scalar_coerce('?int', 0) == 0
    assert scalar_coerce('?int', '0') == 0
    x = symbol('x', '?int')
    assert scalar_coerce('?int', x) is x


def test_scalar_name_dtype():
    x = symbol('x', 'int64')
    assert x._name == 'x'
    assert x.dshape == dshape('int64')


def test_scalar_field():
    x = symbol('x', '{name: string, amount: int64, when: datetime}')
    assert 'amount' in dir(x)
    assert x.amount.dshape == dshape('int64')


def test_scalar_projection():
    x = symbol('x', '{name: string, amount: int64, when: datetime}')
    assert x[['amount', 'when']].dshape == \
            dshape('{amount: int64, when: datetime}')



from blaze.expr import by, symbol

def test_by_raises_informative_error_on_old_syntax():
    s = symbol('t', 'var * {name: string, amount: int}')
    try:
        by(s.name, s.amount.sum())
        assert False
    except ValueError as e:
        assert 'please' in str(e).lower()



from __future__ import absolute_import, division, print_function

import pytest
import operator

from datashape import discover, dshape

from blaze.compute.core import (compute_up, compute, bottom_up_until_type_break,
                                top_then_bottom_then_top_again_etc,
                                swap_resources_into_scope)
from blaze.expr import by, symbol, Expr, Symbol
from blaze.dispatch import dispatch
from blaze.compatibility import raises, reduce
from blaze.utils import example
from blaze.interactive import into

import pandas as pd
import numpy as np
import dask.array as da


def test_errors():
    t = symbol('t', 'var * {foo: int}')
    with raises(NotImplementedError):
        compute_up(by(t, count=t.count()), 1)


def test_optimize():
    class Foo(object):
        pass

    s = symbol('s', '5 * {x: int, y: int}')

    @dispatch(Expr, Foo)
    def compute_down(expr, foo):
        return str(expr)

    assert compute(s.x * 2, Foo()) == "s.x * 2"

    @dispatch(Expr, Foo)
    def optimize(expr, foo):
        return expr + 1

    assert compute(s.x * 2, Foo()) == "(s.x * 2) + 1"


def test_bottom_up_until_type_break():

    s = symbol('s', 'var * {name: string, amount: int}')
    data = np.array([('Alice', 100), ('Bob', 200), ('Charlie', 300)],
                    dtype=[('name', 'S7'), ('amount', 'i4')])

    e = (s.amount + 1).distinct()
    expr, scope = bottom_up_until_type_break(e, {s: data})
    amount = symbol('amount', 'var * int64', token=expr._token)
    assert expr.isidentical(amount)
    assert len(scope) == 1
    assert amount in scope
    assert (scope[amount] == np.array([101, 201, 301], dtype='i4')).all()

    # This computation has a type change midstream, so we stop and get the
    # unfinished computation.

    e = s.amount.sum() + 1
    expr, scope = bottom_up_until_type_break(e, {s: data})
    amount_sum = symbol('amount_sum', 'int64', token=expr.lhs._token)
    assert expr.isidentical(amount_sum + 1)
    assert len(scope) == 1
    assert amount_sum in scope
    assert scope[amount_sum] == 600

    # ensure that we work on binops with one child
    x = symbol('x', 'real')
    expr, scope = bottom_up_until_type_break(x + x, {x: 1})
    assert len(scope) == 1
    x2 = list(scope.keys())[0]
    assert isinstance(x2, Symbol)
    assert isinstance(expr, Symbol)
    assert scope[x2] == 2


def test_top_then_bottom_then_top_again_etc():
    s = symbol('s', 'var * {name: string, amount: int32}')
    data = np.array([('Alice', 100), ('Bob', 200), ('Charlie', 300)],
                    dtype=[('name', 'S7'), ('amount', 'i4')])

    e = s.amount.sum() + 1
    assert top_then_bottom_then_top_again_etc(e, {s: data}) == 601


def test_swap_resources_into_scope():

    from blaze import data
    t = data([1, 2, 3], dshape='3 * int', name='t')
    expr, scope = swap_resources_into_scope(t.head(2), {t: t.data})

    assert t._resources()
    assert not expr._resources()

    assert t not in scope


def test_compute_up_on_dict():
    d = {'a': [1, 2, 3], 'b': [4, 5, 6]}

    assert str(discover(d)) == str(dshape('{a: 3 * int64, b: 3 * int64}'))

    s = symbol('s', discover(d))

    assert compute(s.a, {s: d}) == [1, 2, 3]


def test_pre_compute_on_multiple_datasets_is_selective():
    from odo import CSV
    from blaze import data
    from blaze.cached import CachedDataset

    df = pd.DataFrame([[1, 'Alice',   100],
                         [2, 'Bob',    -200],
                         [3, 'Charlie', 300],
                         [4, 'Denis',   400],
                         [5, 'Edith',  -500]], columns=['id', 'name', 'amount'])
    iris = CSV(example('iris.csv'))
    dset = CachedDataset({'df': df, 'iris': iris})

    d = data(dset)
    assert str(compute(d.df.amount)) == str(df.amount)


def test_raises_on_valid_expression_but_no_implementation():
    class MyExpr(Expr):
        __slots__ = '_hash', '_child'

        @property
        def dshape(self):
            return self._child.dshape

    t = symbol('t', 'var * {amount: real}')
    expr = MyExpr(t.amount)
    df = [(1.0,), (2.0,), (3.0,)]
    with pytest.raises(NotImplementedError):
        compute(expr, df)


@pytest.mark.parametrize('n', range(2, 11))
def test_simple_add(n):
    x = symbol('x', 'int')
    expr = reduce(operator.add, [x] * n)
    assert compute(expr, 1) == n


@pytest.mark.parametrize('data,expr,ret_type,exp_type', [
    (1, symbol('x', 'int'), 'native', int),
    (1, symbol('x', 'int'), 'core', int),
    # use dask array to test core since isn't core type
    (into(da.core.Array, [1, 2], chunks=(10,)), symbol('x', '2 * int'), 'core', pd.Series),  # test 1-d to series
    (into(da.core.Array, [{'a': 1, 'b': 2}, {'a': 3, 'b': 4}], chunks=(10,)), symbol('x', '2 * {a: int, b: int}'), 'core', pd.DataFrame),  # test 2-d tabular to dataframe
    (into(da.core.Array, [[1, 2], [3, 4]], chunks=(10, 10)), symbol('x', '2 *  2 * int'), 'core', np.ndarray),  # test 2-d non tabular to ndarray
    ([1, 2], symbol('x', '2 * int') , tuple, tuple)
])
def test_compute_return_type(data, expr, ret_type, exp_type):
    assert isinstance(compute(expr, data, return_type=ret_type), exp_type)
